"""
define classes to describe space group
"""
__author__ = 'ikibalin'
__version__ = "2019_08_27"
import os
import numpy
from pycifstar import Global

from cryspy.f_common.cl_fitable import Fitable

class SpaceGroup(object):
    """
    Space Group
    """
    def __init__(self, spgr_given_name = "P1", spgr_choice = "1",
                 f_dir_prog = os.path.dirname(__file__)):
        #         f_dir_prog = os.getcwd()):
        super(SpaceGroup, self).__init__()
        
        self.__spgr_given_name = None
        self.__spgr_choice = None
        self.__f_dir_prog = None
        self.__spgr_table = None


        self.__singony = "Triclinic"

        self.__centr = None
        self.__el_symm = None
        self.__orig = None
        self.__p_centr = None
        self.__spgr_name = None
        self.__spgr_number = None

        self.__r_11 = None
        self.__r_12 = None
        self.__r_13 = None
        self.__r_21 = None
        self.__r_22 = None
        self.__r_23 = None
        self.__r_31 = None
        self.__r_32 = None
        self.__r_33 = None

        self.__b_1 = None
        self.__b_2 = None
        self.__b_3 = None

        self.f_dir_prog = f_dir_prog
        self.spgr_name = spgr_given_name
        self.spgr_choice = spgr_choice


    @property
    def spgr_given_name(self):
        """

        reference:
        """
        return self.__spgr_given_name
    @spgr_given_name.setter
    def spgr_given_name(self, x):
        self.__spgr_given_name = "".join(x.split()).strip("\"").strip("'")
        if self.is_defined:
            _ = self._form_object()

    @property
    def spgr_choice(self):
        """

        reference:
        """
        return self.__spgr_choice
    @spgr_choice.setter
    def spgr_choice(self, x):
        if isinstance(x, float):
            x_in = "{:}".format(int(x))
        else:
            x_in = str(x)
        self.__spgr_choice = x_in.strip()
        if self.is_defined:
            _ = self._form_object()

    @property
    def f_dir_prog(self):
        """

        reference:
        """
        return self.__f_dir_prog
    @f_dir_prog.setter
    def f_dir_prog(self, x):
        f_itables = os.path.join(x, "tables", "itables.txt")
        self._read_el_cards(f_itables)    
        self.__f_dir_prog = x
        if self.is_defined:
            _ = self._form_object()



    @property
    def spgr_name(self):
        return self.__spgr_name
    @spgr_name.setter
    def spgr_name(self, x):
        self.__spgr_given_name = "".join(str(x).split()) # __spgr_given_name is not mistake
        if self.is_defined:
            _ = self._form_object()

    @property
    def spgr_number(self):
        return self.__spgr_number
    @property
    def singony(self):
        return self.__singony
    @property
    def bravais_lattice(self):
        return self.__singony
    @property
    def centr(self):
        return self.__centr
    @property
    def el_symm(self):
        return self.__el_symm
    @property
    def orig(self):
        return self.__orig
    @property
    def p_centr(self):
        return self.__p_centr
    @property
    def r_11(self):
        return self.__r_11
    @property
    def r_22(self):
        return self.__r_22
    @property
    def r_33(self):
        return self.__r_33
    @property
    def r_12(self):
        return self.__r_12
    @property
    def r_13(self):
        return self.__r_13
    @property
    def r_23(self):
        return self.__r_23
    @property
    def r_21(self):
        return self.__r_21
    @property
    def r_31(self):
        return self.__r_31
    @property
    def r_32(self):
        return self.__r_32
    @property
    def b_1(self):
        return self.__b_1
    @property
    def b_2(self):
        return self.__b_2
    @property
    def b_3(self):
        return self.__b_3



    def _show_message(self, s_out: str):
        print("***  Error ***")
        print(s_out)

    def __repr__(self):
        ls_out = ["SpaceGroup:\n given name: {:}\n choice: {:}".format(self.spgr_given_name, self.spgr_choice)]
        if self.spgr_name is not None:
            ls_out.append(" name: {:}".format(self.spgr_name))
        if self.spgr_number is not None:
            ls_out.append(" number: {:}".format(self.spgr_number))
        #ls_out.append(" {:}\n directory: '{:}'".format(self._trans_el_symm_to_str(), self.f_dir_prog))
        return "\n".join(ls_out)

    @property
    def is_defined(self):
        """
        Output: True if all started parameters are given
        """
        cond = any([self.spgr_given_name is None, 
                    self.spgr_choice is None, 
                    self.f_dir_prog is None])
        return not(cond)
        
    def _read_el_cards(self, f_itables):
        """
        reading information about space grooupe from file fitables to list of cards ldcard
        Info in file fitables:
        
        1 P1               Triclinic
        choice: 1
        centr: false
        pcentr: 0, 0, 0
        symmetry: X,Y,Z
        
        2 P-1              Triclinic
        ...
        """
        fid = open(f_itables, "r")
        lcontent = fid.readlines()
        fid.close()
    
        lcontent = [hh.strip() for hh in lcontent if hh.strip() != ""]
        ldcard = []
        dcard = None
        for hh in lcontent:
            lhelp = hh.split()
            if lhelp[0].isdigit():
                if dcard != None:
                    ldcard.append(dcard)
                dcard = {"number":lhelp[0], "name": lhelp[1], "singony": lhelp[2]}
            else:
                lhelp = hh.split(":")
                if (lhelp[0].strip() in dcard.keys()):
                    dcard[lhelp[0].strip()].append(lhelp[1].strip())
                else:
                    dcard[lhelp[0].strip()] = [lhelp[1].strip()]
        ldcard.append(dcard)
        self._p_spgr_table = ldcard
        

    def _get_symm(self):
        """
        get symmetry from space group
        """
        
        spgr_choice = self.spgr_choice
        spgr_given_name = self.spgr_given_name
        

        if spgr_given_name.isdigit():
            spgr_n = spgr_given_name
            spgr_name = ""
        else:
            spgr_n = ""
            spgr_name = spgr_given_name
        
        spgr_table = self._p_spgr_table
        flag = False
        for dcard in spgr_table:
            if (((dcard["number"] == spgr_n)|(dcard["name"] == spgr_name))&(dcard["choice"][0] == spgr_choice)):
                flag = True
                break
        if (not flag):
            print("Space group is not found: {:} {:} {:}".format(spgr_n, spgr_name, spgr_choice))
            return
        
        flag = False
            
        l_el_symm = []
        for ssymm in dcard["symmetry"]:
            l_el_symm.append(self._trans_str_to_el_symm(ssymm))
        centr = dcard["centr"][0]=="true"
        pcentr = [float(hh) for hh in dcard["pcentr"][0].split(",")]
        fletter = dcard["name"][0]
        spgr = dcard["name"]
        number = int(dcard["number"])
        singony = dcard["singony"]
        if (fletter == "P"):
            l_orig = [(0, 0, 0)]
        elif fletter == "C":
            l_orig = [(0, 0, 0), (0.5, 0.5, 0)]
        elif fletter == "I":
            l_orig = [(0, 0, 0), (0.5, 0.5, 0.5)]
        elif fletter == "F":
            l_orig = [(0, 0, 0), (0.5, 0.5, 0), (0.5, 0, 0.5), (0, 0.5, 0.5)]
        elif (fletter == "R"):
            if spgr_choice == "1":
                l_orig = [(0, 0, 0), (0.66667, 0.33333, 0.33333), (0.33334, 0.66666, 0.66666)]
            else:
                l_orig = [(0, 0, 0)]
        else:
            print("Undefined syngony")

            
        self.__centr = centr
        self.__el_symm = l_el_symm
        self.__orig = l_orig
        self.__p_centr = pcentr
        self.__spgr_name = spgr
        self.__spgr_number = number
        self.__singony = singony
        
    def _calc_rotation_matrix_anb_b(self):
        """
        give representation for rotation matrix: r_11, r_22, r_33, r_12, r_13, r_23 and vector b_1, b_2, b_3
        """
        lel_symm = self.el_symm
        b_1 = numpy.array([hh[0] for hh in lel_symm], dtype = float)
        r_11 = numpy.array([hh[1] for hh in lel_symm], dtype = int)
        r_12 = numpy.array([hh[2] for hh in lel_symm], dtype = int)
        r_13 = numpy.array([hh[3] for hh in lel_symm], dtype = int)

        b_2 = numpy.array([hh[4] for hh in lel_symm], dtype = float)
        r_21 = numpy.array([hh[5] for hh in lel_symm], dtype = int)
        r_22 = numpy.array([hh[6] for hh in lel_symm], dtype = int)
        r_23 = numpy.array([hh[7] for hh in lel_symm], dtype = int)

        b_3 = numpy.array([hh[8] for hh in lel_symm], dtype = float)
        r_31 = numpy.array([hh[9] for hh in lel_symm], dtype = int)
        r_32 = numpy.array([hh[10] for hh in lel_symm], dtype = int)
        r_33 = numpy.array([hh[11] for hh in lel_symm], dtype = int)
        
        self.__r_11 = r_11
        self.__r_12 = r_12
        self.__r_13 = r_13

        self.__r_21 = r_21
        self.__r_22 = r_22
        self.__r_23 = r_23

        self.__r_31 = r_31
        self.__r_32 = r_32
        self.__r_33 = r_33

        self.__b_1 = b_1
        self.__b_2 = b_2
        self.__b_3 = b_3

    def _form_object(self):
        if self.is_defined:
            self._get_symm()
            self._calc_rotation_matrix_anb_b()
        else:
            self._show_message("Object is not properly defined")
            return False
        return True

    def calc_hkl_equiv(self, h, k, l):
        """
        give equivalent reflections of hkl and its multiplicity
        """

        r_11 = self.r_11
        r_12 = self.r_12
        r_13 = self.r_13
        r_21 = self.r_21
        r_22 = self.r_22
        r_23 = self.r_23
        r_31 = self.r_31
        r_32 = self.r_32
        r_33 = self.r_33

        h_s = r_11*h + r_21*k + r_31*l 
        k_s = r_12*h + r_22*k + r_32*l 
        l_s = r_13*h + r_23*k + r_33*l 
        
        hkl_s = numpy.vstack([h_s, k_s, l_s])
        hkl_s = numpy.hstack([hkl_s,-1*hkl_s])
        hkl_s_un = numpy.unique(hkl_s, axis=1)
        multiplicity = int(round(hkl_s.shape[1]*1./hkl_s_un.shape[1]))
        h_s, k_s, l_s = hkl_s_un[0, :], hkl_s_un[1, :], hkl_s_un[2, :]
        return h_s, k_s, l_s, multiplicity

    def calc_xyz_mult(self, x, y, z):
        """
        give unique x,y,z elements and calculate multiplicit for given x,y,z fract
        """
        r_11 = self.r_11
        r_12 = self.r_12
        r_13 = self.r_13
        r_21 = self.r_21
        r_22 = self.r_22
        r_23 = self.r_23
        r_31 = self.r_31
        r_32 = self.r_32
        r_33 = self.r_33
        b_1 = self.b_1
        b_2 = self.b_2
        b_3 = self.b_3
        
        l_orig = self.orig
        centr = self.centr
        p_centr = self.p_centr

        x_s = numpy.round(numpy.mod(r_11*x + r_12*y + r_13*z + b_1, 1), 6)
        y_s = numpy.round(numpy.mod(r_21*x + r_22*y + r_23*z + b_2, 1), 6)
        z_s = numpy.round(numpy.mod(r_31*x + r_32*y + r_33*z + b_3, 1), 6)

        x_o = [orig[0] for orig in l_orig]
        y_o = [orig[1] for orig in l_orig]
        z_o = [orig[2] for orig in l_orig]
        
        x_s_2d, x_o_2d = numpy.meshgrid(x_s, x_o)
        y_s_2d, y_o_2d = numpy.meshgrid(y_s, y_o)
        z_s_2d, z_o_2d = numpy.meshgrid(z_s, z_o)
        
        x_s_2d = numpy.round(numpy.mod(x_s_2d+x_o_2d, 1), 6)
        y_s_2d = numpy.round(numpy.mod(y_s_2d+y_o_2d, 1), 6)
        z_s_2d = numpy.round(numpy.mod(z_s_2d+z_o_2d, 1), 6)

        x_s = x_s_2d.flatten()
        y_s = y_s_2d.flatten()
        z_s = z_s_2d.flatten()

        if centr:
            x_s_h = numpy.round(numpy.mod(2.*p_centr[0]-1.*x_s, 1), 6)
            y_s_h = numpy.round(numpy.mod(2.*p_centr[1]-1.*y_s, 1), 6)
            z_s_h = numpy.round(numpy.mod(2.*p_centr[2]-1.*z_s, 1), 6)
            x_s =numpy.hstack([x_s, x_s_h])
            y_s =numpy.hstack([y_s, y_s_h])
            z_s =numpy.hstack([z_s, z_s_h])
                        
        xyz_s = numpy.vstack([x_s, y_s, z_s])

        xyz_s_un = numpy.unique(xyz_s, axis=1)
        n_atom = int(round(xyz_s.shape[1]*1./xyz_s_un.shape[1]))
        x_s, y_s, z_s = xyz_s_un[0, :], xyz_s_un[1, :], xyz_s_un[2, :]
        
        return x_s, y_s, z_s, n_atom
    
    def calc_el_symm_for_xyz(self, x_in, y_in, z_in):
        x ,y, z = x_in%1., y_in%1., z_in%1.
        r_11 = self.r_11
        r_12 = self.r_12
        r_13 = self.r_13
        r_21 = self.r_21
        r_22 = self.r_22
        r_23 = self.r_23
        r_31 = self.r_31
        r_32 = self.r_32
        r_33 = self.r_33
        b_1 = self.b_1
        b_2 = self.b_2
        b_3 = self.b_3
        
        l_orig = self.orig
        centr = self.centr
        p_centr = self.p_centr


        x_o = [orig[0] for orig in l_orig]
        y_o = [orig[1] for orig in l_orig]
        z_o = [orig[2] for orig in l_orig]
        
        r_11_2d, x_o_2d = numpy.meshgrid(r_11, x_o, indexing="ij")
        r_12_2d, y_o_2d = numpy.meshgrid(r_12, y_o, indexing="ij")
        r_13_2d, z_o_2d = numpy.meshgrid(r_13, z_o, indexing="ij")

        r_21_2d = numpy.meshgrid(r_21, x_o, indexing="ij")[0]
        r_22_2d = numpy.meshgrid(r_22, y_o, indexing="ij")[0]
        r_23_2d = numpy.meshgrid(r_23, z_o, indexing="ij")[0]
        
        r_31_2d = numpy.meshgrid(r_31, x_o, indexing="ij")[0]
        r_32_2d = numpy.meshgrid(r_32, y_o, indexing="ij")[0]
        r_33_2d = numpy.meshgrid(r_33, z_o, indexing="ij")[0]
        
        b_1_2d = numpy.meshgrid(b_1, x_o, indexing="ij")[0]
        b_2_2d = numpy.meshgrid(b_2, y_o, indexing="ij")[0]
        b_3_2d = numpy.meshgrid(b_3, z_o, indexing="ij")[0]
        
        b_1_2d = b_1_2d + x_o_2d
        b_2_2d = b_2_2d + x_o_2d
        b_3_2d = b_3_2d + x_o_2d

        e_11 = r_11_2d.flatten()
        e_12 = r_12_2d.flatten()
        e_13 = r_13_2d.flatten()

        e_21 = r_21_2d.flatten()
        e_22 = r_22_2d.flatten()
        e_23 = r_23_2d.flatten()

        e_31 = r_31_2d.flatten()
        e_32 = r_32_2d.flatten()
        e_33 = r_33_2d.flatten()

        e_1 = b_1_2d.flatten()
        e_2 = b_2_2d.flatten()
        e_3 = b_3_2d.flatten()

        if centr:
            me_11, me_12, me_13 = -1*e_11, -1*e_12, -1*e_13
            me_21, me_22, me_23 = -1*e_21, -1*e_22, -1*e_23
            me_31, me_32, me_33 = -1*e_31, -1*e_32, -1*e_33
            me_1 = 2.*p_centr[0]-1.*e_1
            me_2 = 2.*p_centr[1]-1.*e_2
            me_3 = 2.*p_centr[2]-1.*e_3 
            
            e_11 = numpy.hstack([e_11, me_11])
            e_12 = numpy.hstack([e_12, me_12])
            e_13 = numpy.hstack([e_13, me_13])
                                          
            e_21 = numpy.hstack([e_21, me_21])
            e_22 = numpy.hstack([e_22, me_22])
            e_23 = numpy.hstack([e_23, me_23])
                                          
            e_31 = numpy.hstack([e_31, me_31])
            e_32 = numpy.hstack([e_32, me_32])
            e_33 = numpy.hstack([e_33, me_33])

            e_1 = numpy.hstack([e_1, me_1])
            e_2 = numpy.hstack([e_2, me_2])
            e_3 = numpy.hstack([e_3, me_3])

            
        x_s = numpy.round(numpy.mod(e_11*x + e_12*y + e_13*z + e_1, 1), 5)
        y_s = numpy.round(numpy.mod(e_21*x + e_22*y + e_23*z + e_2, 1), 5)
        z_s = numpy.round(numpy.mod(e_31*x + e_32*y + e_33*z + e_3, 1), 5)
            
        xyz_s = numpy.vstack([x_s, y_s, z_s])
        
        xyz_s_un, unique_inverse = numpy.unique(xyz_s, return_inverse=True, axis=1)
        x_s, y_s, z_s = xyz_s_un[0, :], xyz_s_un[1, :], xyz_s_un[2, :]
        ind = (numpy.where((x-x_s)**2+(y-y_s)**2+(z-z_s)**2 < 0.00001))[0][0]

        flag = unique_inverse == ind
        o_11, o_12, o_13 = e_11[flag], e_12[flag], e_13[flag]
        o_21, o_22, o_23 = e_21[flag], e_22[flag], e_23[flag]
        o_31, o_32, o_33 = e_31[flag], e_32[flag], e_33[flag]
        o_3, o_2, o_3 = e_3[flag], e_2[flag], e_3[flag]
        return o_11, o_12, o_13, o_21, o_22, o_23, o_31, o_32, o_33, o_3, o_2, o_3


    
    def calc_atom_mult(self, np_x, np_y, np_z):
        """
        calculate atom multiplicity
        """
        lmult=[]
        for x, y, z in zip(np_x, np_y, np_z):
            np_x_s = self.calc_xyz_mult(x, y, z)[0]
            lmult.append(np_x_s.shape[0])
        np_multiplicity = numpy.array(lmult, dtype=int)
        
        return np_multiplicity
    
    def _trans_str_to_el_symm(self, str1):
        """
        transform string to element of symmetry: (x,y,-z) -> 0.0 1 0 0  0.0 0 1 0  0.0 0 0 -1
        """
        str2 = "".join(str1.split(" "))
        l_help1, l_help2, l_help3 = [], [], []
        l_help1 = [hh for hh in str2.split('(') if hh != ""]
        [l_help2.extend(hh.split(')')) for hh in l_help1 if hh != ""]
        [l_help3.extend(hh.split(',')) for hh in l_help2 if hh != ""]
        l_Ax = ['x', 'y', 'z']
        l_el_symm = []
        for hh in l_help3:
            el_symm_h = [0.0, 0, 0, 0]
            str_h = hh
            for i_num, Ax in enumerate(l_Ax):
                if (str_h.find(Ax) != -1):
                    if (str_h.find("+"+Ax) != -1):
                        el_symm_h[i_num+1] = 1
                        str_h = "".join(str_h.split("+"+Ax))
                    elif (str_h.find("-"+Ax) != -1):
                        el_symm_h[i_num+1] = -1
                        str_h = "".join(str_h.split("-"+Ax))
                    else:
                        el_symm_h[i_num+1] = 1
                        str_h = "".join(str_h.split(Ax))
            if (str_h==""):
                pass
            elif (str_h.find("/") != -1):
                l_help1 = str_h.split("/")
                el_symm_h[0] = float(l_help1[0])/float(l_help1[1])
            else:
                el_symm_h[0] = float(str_h)
            l_el_symm.append(el_symm_h)
        el_symm = []
        [el_symm.extend(hh) for hh in l_el_symm]
        return el_symm

    def _trans_el_symm_to_str(self):
        ls_out = []
        l_el_symm = self.el_symm
        centr = self.centr
        if centr:
            p_centr = self.p_centr
            ls_out.append("inversion center at ({:.3f}, {:.3f}, {:.3f})".format(
                    p_centr[0], p_centr[1], p_centr[2]))
        if l_el_symm == []:
            return ""
        for el_symm in l_el_symm:
            s_x = ""
            if el_symm[0] != 0.: s_x+="{:.3f}".format(el_symm[0])
            if el_symm[1] == 1: s_x+="+x"
            if el_symm[1] == -1: s_x+="-x"
            if el_symm[2] == 1: s_x+="+y"
            if el_symm[2] == -1: s_x+="-y"
            if el_symm[3] == 1: s_x+="+z"
            if el_symm[3] == -1: s_x+="-z"
            if s_x.startswith("+"): s_x = s_x[1:]
            
            s_y = ""
            if el_symm[4] != 0.: s_y+="{:.3f}".format(el_symm[4])
            if el_symm[5] == 1: s_y+="+x"
            if el_symm[5] == -1: s_y+="-x"
            if el_symm[6] == 1: s_y+="+y"
            if el_symm[6] == -1: s_y+="-y"
            if el_symm[7] == 1: s_y+="+z"
            if el_symm[7] == -1: s_y+="-z"
            if s_y.startswith("+"): s_y = s_y[1:]

            s_z = ""
            if el_symm[8] != 0.: s_z+="{:.3f}".format(el_symm[8])
            if el_symm[9]==1: s_z+="+x"
            if el_symm[9] == -1: s_z+="-x"
            if el_symm[10] == 1: s_z+="+y"
            if el_symm[10] == -1: s_z+="-y"
            if el_symm[11] == 1: s_z+="+z"
            if el_symm[11] == -1: s_z+="-z"
            if s_z.startswith("+"): s_z = s_z[1:]

            line=" {:}, {:}, {:}".format(s_x, s_y, s_z)
            ls_out.append(line)
        return "\n".join(ls_out)

    def calc_asymmetric_cell(self, n_a, n_b, n_c):
        """
        give the numbers in asymmetric cell

        n_a is the number of points along a axis
        n_b is the numper of points along b axis
        n_c is the numper of points along c axis

        na, n_b, nc should be divided on 24: 8 and 3

        output: 
        l_coord is a list of coordinates in asymmetric cell (frac_x = n_x/n_a and so on)
        l_symm contains a list of symmetry given as (n_symm, centr, n_orig)
        """

        n_a_new = int(round(n_a/24))*24
        n_b_new = int(round(n_b/24))*24
        n_c_new = int(round(n_c/24))*24

        #print("na: {:}, n_b: {:}, n_c: {:}".format(n_a_new, n_b_new, n_c_new))

        l_el_symm = self.el_symm
        f_centr = self.centr
        p_centr = self.p_centr
        l_orig = self.orig
        l_coord = []
        
        spgr_choice = self.spgr_choice 
        spgr_name = self.spgr_name 
        spgr_number = self.spgr_number
        
        if (spgr_number==227) & (spgr_choice=="2"):
            n_a_new = int(round(n_a/8))*8
            for n_x in range(-n_a_new//8, 3*n_a_new//8+1):
                for n_y in range(-n_a_new//8, 0+1):
                    for n_z in range(-n_a_new//4, 0+1):
                        cond_1 = (n_y < min([n_a_new//4-n_x, n_x]))
                        cond_2 = (n_z >= -n_y-n_a_new//4)
                        cond_3 = (n_z <= n_y)
                        if (cond_1 & cond_2 & cond_3):
                            coord_x, coord_y = float(n_x)/float(n_a_new), float(n_y)/float(n_a_new)
                            coord_z = float(n_z)/float(n_a_new)
                            #print(" {:3} {:3} {:3}".format(n_x, n_y, n_z), " ", " {:6.3f} {:6.3f} {:6.3f}".format(coord_x, coord_y, coord_z))
                            l_coord.append((coord_x, coord_y, coord_z))
        return l_coord

    def from_cif(self, string: str):
        cif_global = Global()
        flag = cif_global.take_from_string(string)
        if not flag:
            return False
        flag = False
        if cif_global.is_value("_space_group_name_H-M_alt"):
            self.spgr_name = cif_global["_space_group_name_H-M_alt"].value
        if cif_global.is_value("_space_group_it_coordinate_system_code"):
            self.spgr_choice = cif_global["_space_group_it_coordinate_system_code"].value
        return True

    @property
    def to_cif(self):
        ls_out = []
        ls_out.append("_space_group_name_H-M_alt {:}".format(self.spgr_name))
        ls_out.append("_space_group_it_coordinate_system_code {:}".format(self.spgr_choice))
        return "\n".join(ls_out)



#from cryspy.f_experiment.f_powder_2d.cl_asymmetry_powder_2d import AsymmetryPowder2D
#from cryspy.f_experiment.f_powder_2d.cl_background_powder_2d import BackgroundPowder2D
#from cryspy.f_experiment.f_powder_2d.cl_calculated_data_powder_2d import CalculatedDataPowder2D
#from cryspy.f_experiment.f_powder_2d.cl_experiment_powder_2d import ExperimentPowder2D
#from cryspy.f_experiment.f_powder_2d.cl_factor_lorentz_powder_2d import FactorLorentzPowder2D
#from cryspy.f_experiment.f_powder_2d.cl_observed_data_powder_2d import ObservedDataPowder2D
#from cryspy.f_experiment.f_powder_2d.cl_resolution_powder_2d import ResolutionPowder2D
#from cryspy.f_experiment.f_powder_2d.cl_setup_powder_2d import SetupPowder2D

from .cl_pd2d_background import Pd2dBackground
from .cl_pd2d_exclude import Pd2dExclude
from .cl_pd2d_instr_reflex_asymmetry import Pd2dInstrReflexAsymmetry
from .cl_pd2d_instr_resolution import Pd2dInstrResolution
from .cl_pd2d_meas import Pd2dMeas
from .cl_pd2d_phase import Pd2dPhase
from .cl_pd2d_peak import Pd2dPeak
from .cl_pd2d_proc import Pd2dProc
from .cl_pd2d import Pd2d







"""
define classe Pd2d which describes the 2d powder diffraction experiment
"""
__author__ = 'ikibalin'
__version__ = "2019_09_11"
import os
import numpy
from pycifstar import Data, Loop

from cryspy.f_common.cl_fitable import Fitable
from cryspy.f_experiment.cl_beam_polarization import BeamPolarization
from .cl_pd2d_background import Pd2dBackground
from .cl_pd2d_exclude import Pd2dExclude
from .cl_pd2d_instr_reflex_asymmetry import Pd2dInstrReflexAsymmetry
from .cl_pd2d_instr_resolution import Pd2dInstrResolution
from .cl_pd2d_meas import Pd2dMeas
from .cl_pd2d_phase import Pd2dPhase
from .cl_pd2d_proc import Pd2dProc
from .cl_pd2d_peak import Pd2dPeak


from cryspy.f_crystal.cl_magnetism import calc_mRmCmRT

class Pd2d(object):
    """
    Class to describe information about single diffraction measurements

    Example:

    
    _diffrn_radiation_wavelength 1.40 
    _diffrn_ambient_field  1.000

    _pd2d_chi2_sum True
    _pd2d_chi2_diff False
    _pd2d_chi2_up False
    _pd2d_chi2_down False

    _pd2d_calib_2theta_offset -0.385404

    _pd2d_2theta_range_min 4.0
    _pd2d_2theta_range_max 80.0
    _pd2d_phi_range_min -2.0
    _pd2d_phi_range_max 40.0
    """
    def __init__(self, label="powder2", 
                 background = None, exclude=Pd2dExclude(ttheta_min=[0.], ttheta_max=[1.], phi_min=[0.0], phi_max=[0.1]), 
                 asymmetry=Pd2dInstrReflexAsymmetry(),
                 beam_polarization = BeamPolarization(), resolution = Pd2dInstrResolution(), meas=Pd2dMeas(), phase=Pd2dPhase(),
                 wavelength=1.4, field=1.0,
                 chi2_sum = True, chi2_diff = True, chi2_up = False, chi2_down = False,
                 offset = Fitable(0.), ttheta_min = 0.1, ttheta_max = 179.9, phi_min=0., phi_max=40.):
        super(Pd2d, self).__init__()
        self.__label = None
        self.__background = None
        self.__exclude_2theta = None
        self.__asymmetry = None
        self.__beam_polarization = None
        self.__resolution = None
        self.__meas = None 
        self.__phase = None 
        self.__proc = None 
        self.__diffrn_radiation_wavelength = None 
        self.__diffrn_ambient_field = None 
        self.__pd2d_chi2_sum = None 
        self.__pd2d_chi2_diff = None 
        self.__pd2d_chi2_up = None 
        self.__pd2d_chi2_down = None 
        self.__pd2d_calib_2theta_offset = None 
        self.__pd2d_2theta_range_min = None 
        self.__pd2d_2theta_range_max = None 
        self.__pd2d_phi_range_min = None 
        self.__pd2d_phi_range_max = None 

        self.__dd = None 

        self.label = label
        self.background = background
        self.exclude = exclude
        self.asymmetry = asymmetry
        self.beam_polarization = beam_polarization
        self.resolution = resolution
        self.meas = meas
        self.phase = phase
        self.wavelength = wavelength
        self.field = field
        self.chi2_sum = chi2_sum
        self.chi2_diff = chi2_diff
        self.chi2_up = chi2_up
        self.chi2_down = chi2_down
        self.offset = offset
        self.ttheta_max = ttheta_max
        self.ttheta_min = ttheta_min
        self.phi_max = phi_max
        self.phi_min = phi_min

        self.__proc = None
        self.__peaks = None
        self.__reflns = None

    @property
    def erase_precalc(self):
        """
        Erase precalculated values
        """
        self.__proc = None
        self.__peaks = None
        self.__reflns = None

    @property
    def label(self):
        return self.__label
    @label.setter
    def label(self, x: str):
        self.__label = str(x)

    @property
    def background(self):
        return self.__background
    @background.setter
    def background(self, x):
        if isinstance(x, Pd2dBackground):
            x_in = x
        elif isinstance(x, str):
            x_in = Pd2dBackground()
            flag = x_in.from_cif(x)
            if not(flag):
                x_in = None
                self._show_message("A induced string can not be converted to Pd2dBackground")
        elif x is None:
            x_in = None
        else:
            x_in = None
            self._show_message("Input type for Pd2dBackground is not recognized")
        self.__background = x_in

    @property
    def exclude(self):
        return self.__exclude_2theta
    @exclude.setter
    def exclude(self, x):
        if isinstance(x, Pd2dExclude):
            x_in = x
        elif isinstance(x, str):
            x_in = Pd2dExclude()
            flag = x_in.from_cif(x)
            if not(flag):
                x_in = None
                self._show_message("A induced string can not be converted to Pd2dExclude")
        elif x is None:
            x_in = None
        else:
            x_in = None
            self._show_message("Input type for Pd2dExclude is not recognized")
        self.__exclude_2theta = x_in
        self.erase_precalc

    @property
    def asymmetry(self):
        return self.__asymmetry
    @asymmetry.setter
    def asymmetry(self, x):
        if isinstance(x, Pd2dInstrReflexAsymmetry):
            x_in = x
        elif isinstance(x, str):
            x_in = Pd2dInstrReflexAsymmetry()
            flag = x_in.from_cif(x)
            if not(flag):
                x_in = None
                self._show_message("A induced string can not be converted to Pd2dAsymmetry")
        elif x is None:
            x_in = None
        else:
            x_in = None
            self._show_message("Input type for Pd2dAsymmetry is not recognized")
        self.__asymmetry = x_in

    @property
    def beam_polarization(self):
        return self.__beam_polarization
    @beam_polarization.setter
    def beam_polarization(self, x):
        if isinstance(x, BeamPolarization):
            x_in = x
        elif isinstance(x, str):
            x_in = BeamPolarization()
            flag = x_in.from_cif(x)
            if not(flag):
                x_in = None
                self._show_message("A induced string can not be converted to BeamPolarization")
        elif x is None:
            x_in = None
        else:
            x_in = None
            self._show_message("Input type for BeamPolarization is not recognized")
        self.__beam_polarization = x_in

    @property
    def resolution(self):
        return self.__resolution
    @resolution.setter
    def resolution(self, x):
        if isinstance(x, Pd2dInstrResolution):
            x_in = x
        elif isinstance(x, str):
            x_in = Pd2dInstrResolution()
            flag = x_in.from_cif(x)
            if not(flag):
                x_in = None
                self._show_message("A induced string can not be converted to Pd2dResolution")
        elif x is None:
            x_in = None
        else:
            x_in = None
            self._show_message("Input type for Pd2dResolution is not recognized")
        self.__resolution = x_in

    @property
    def meas(self):
        return self.__meas
    @meas.setter
    def meas(self, x):
        if isinstance(x, Pd2dMeas):
            x_in = x
        elif isinstance(x, str):
            x_in = Pd2dMeas()
            flag = x_in.from_cif(x)
            if not(flag):
                x_in = None
                self._show_message("A induced string can not be converted to Pd2dMeas")
        elif x is None:
            x_in = None
        else:
            x_in = None
            self._show_message("Input type for Pd2dMeas is not recognized")
        self.__meas = x_in
        self.erase_precalc


    @property
    def phase(self):
        return self.__phase
    @phase.setter
    def phase(self, x):
        if isinstance(x, Pd2dPhase):
            x_in = x
        elif isinstance(x, str):
            x_in = Pd2dPhase()
            flag = x_in.from_cif(x)
            if not(flag):
                x_in = None
                self._show_message("A induced string can not be converted to Pd2dPhase")
        elif x is None:
            x_in = None
        else:
            x_in = None
            self._show_message("Input type for Pd2dPhase is not recognized")
        self.__phase = x_in
        self.erase_precalc



    @property
    def proc(self):
        return self.__proc
    @proc.setter
    def proc(self, x):
        if isinstance(x, Pd2dProc):
            x_in = x
        elif isinstance(x, str):
            x_in = Pd2dProc()
            flag = x_in.from_cif(x)
            if not(flag):
                x_in = None
                self._show_message("A induced string can not be converted to Pd2dProc")
        elif x is None:
            x_in = None
        else:
            x_in = None
            self._show_message("Input type for Pd2dProc is not recognized")
        self.__proc = x_in


    @property
    def wavelength(self):
        return self.__diffrn_radiation_wavelength
    @wavelength.setter
    def wavelength(self, x):
        if isinstance(x, float):
            x_in = x
        elif x is None:
            x_in = None
        else:
            x_in = float(x)
        self.__diffrn_radiation_wavelength = x_in
        self.erase_precalc


    @property
    def field(self):
        return self.__diffrn_ambient_field
    @field.setter
    def field(self, x):
        if isinstance(x, float):
            x_in = x
        elif x is None:
            x_in = None
        else:
            x_in = float(x)
        self.__diffrn_ambient_field = x_in
        self.erase_precalc


    @property
    def ttheta_min(self):
        return self.__pd2d_2theta_range_min
    @ttheta_min.setter
    def ttheta_min(self, x):
        if isinstance(x, float):
            x_in = x
        elif x is None:
            x_in = None
        else:
            x_in = float(x)
        self.__pd2d_2theta_range_min = x_in
        self.erase_precalc


    @property
    def phi_min(self):
        return self.__pd2d_phi_range_min
    @phi_min.setter
    def phi_min(self, x):
        if isinstance(x, float):
            x_in = x
        elif x is None:
            x_in = None
        else:
            x_in = float(x)
        self.__pd2d_phi_range_min = x_in
        self.erase_precalc



    @property
    def ttheta_max(self):
        return self.__pd2d_2theta_range_max
    @ttheta_max.setter
    def ttheta_max(self, x):
        if isinstance(x, float):
            x_in = x
        elif x is None:
            x_in = None
        else:
            x_in = float(x)
        self.__pd2d_2theta_range_max = x_in
        self.erase_precalc



    @property
    def phi_max(self):
        return self.__pd2d_phi_range_max
    @phi_max.setter
    def phi_max(self, x):
        if isinstance(x, float):
            x_in = x
        elif x is None:
            x_in = None
        else:
            x_in = float(x)
        self.__pd2d_phi_range_max = x_in
        self.erase_precalc


    @property
    def chi2_sum(self):
        return self.__pd2d_chi2_sum
    @chi2_sum.setter
    def chi2_sum(self, x):
        if isinstance(x, bool):
            x_in = x
        elif x is None:
            x_in = None
        else:
            x_in = bool(x)
        self.__pd2d_chi2_sum = x_in

    @property
    def chi2_diff(self):
        return self.__pd2d_chi2_diff
    @chi2_diff.setter
    def chi2_diff(self, x):
        if isinstance(x, bool):
            x_in = x
        elif x is None:
            x_in = None
        else:
            x_in = bool(x)
        self.__pd2d_chi2_diff = x_in

    @property
    def chi2_up(self):
        return self.__pd2d_chi2_up
    @chi2_up.setter
    def chi2_up(self, x):
        if isinstance(x, bool):
            x_in = x
        elif x is None:
            x_in = None
        else:
            x_in = bool(x)
        self.__pd2d_chi2_up = x_in

    @property
    def chi2_down(self):
        return self.__pd2d_chi2_down
    @chi2_down.setter
    def chi2_down(self, x):
        if isinstance(x, bool):
            x_in = x
        elif x is None:
            x_in = None
        else:
            x_in = bool(x)
        self.__pd2d_chi2_down = x_in

    @property
    def offset(self):
        return self.__pd2d_calib_2theta_offset
    @offset.setter
    def offset(self, x):
        if isinstance(x, Fitable):
            x_in = x
        else:
            x_in = Fitable()
            flag = x_in.take_it(x)
        self.__pd2d_calib_2theta_offset = x_in

    @property
    def peaks(self):
        return self.__peaks
    @peaks.setter
    def peaks(self, x):
        self.__peaks = x

    @property
    def reflns(self):
        return self.__reflns
    @reflns.setter
    def reflns(self, x):
        self.__reflns = x

    def _show_message(self, s_out: str):
        print("***  Error ***")
        print(s_out)

    def __repr__(self):
        ls_out = ["Pd2d:"]
        if self.label is not None:
            ls_out.append(" label: {:}".format(self.label))
        if self.wavelength is not None:
            ls_out.append(" wavelength: {:.3f}".format(float(self.wavelength)))
        if self.field is not None:
            ls_out.append(" field: {:.3f}".format(float(self.field)))
        if self.ttheta_min is not None:
            ls_out.append(" ttheta_min: {:.3f}".format(float(self.ttheta_min)))
        if self.ttheta_max is not None:
            ls_out.append(" ttheta_max: {:.3f}".format(float(self.ttheta_max)))
        if self.phi_min is not None:
            ls_out.append(" phi_min: {:.3f}".format(float(self.phi_min)))
        if self.phi_max is not None:
            ls_out.append(" phi_max: {:.3f}".format(float(self.phi_max)))
        if self.offset is not None:
            ls_out.append(" offset: {:}".format(self.offset.print_with_sigma))
        if self.chi2_up is not None:
            ls_out.append(" chi2_up: {:}".format(self.chi2_up))
        if self.chi2_down is not None:
            ls_out.append(" chi2_down: {:}".format(self.chi2_down))
        if self.chi2_sum is not None:
            ls_out.append(" chi2_sum: {:}".format(self.chi2_sum))
        if self.chi2_diff is not None:
            ls_out.append(" chi2_diff: {:}".format(self.chi2_diff))

        if self.background is not None:
            ls_out.append("\n"+str(self.background))
        if self.exclude is not None:
            ls_out.append("\n"+str(self.exclude))
        if self.asymmetry is not None:
            ls_out.append("\n"+str(self.asymmetry))
        if self.beam_polarization is not None:
            ls_out.append("\n"+str(self.beam_polarization))
        if self.resolution is not None:
            ls_out.append("\n"+str(self.resolution))
        if self.phase is not None:
            ls_out.append("\n"+str(self.phase))
        if self.reflns is not None:
            ls_out.extend(["\n"+str(_) for _ in self.reflns])
        if self.peaks is not None:
            ls_out.extend(["\n"+str(_) for _ in self.peaks])
        if self.proc is not None:
            ls_out.append("\n"+str(self.proc))
        if self.meas is not None:
            ls_out.append("\n"+str(self.meas))
        return "\n".join(ls_out)

    
    @property
    def is_variable(self):
        l_bool = []
        if self.background is not None:
            l_bool.append(self.background.is_variable)
        if self.exclude is not None:
            l_bool.append(self.exclude.is_variable)
        if self.asymmetry is not None:
            l_bool.append(self.asymmetry.is_variable)
        if self.beam_polarization is not None:
            l_bool.append(self.beam_polarization.is_variable)
        if self.resolution is not None:
            l_bool.append(self.resolution.is_variable)
        if self.phase is not None:
            l_bool.append(self.phase.is_variable)
        l_bool.append(self.is_variable_offset)
        res = any(l_bool)
        return res

    @property
    def is_variable_offset(self):
        l_bool = []
        l_bool.append(self.offset.refinement)
        res = any(l_bool)
        return res

    def get_variables(self):
        l_variable = []
        if self.background is not None:
            l_variable.extend(self.background.get_variables())
        if self.exclude is not None:
            l_variable.extend(self.exclude.get_variables())
        if self.asymmetry is not None:
            l_variable.extend(self.asymmetry.get_variables())
        if self.beam_polarization is not None:
            l_variable.extend(self.beam_polarization.get_variables())
        if self.resolution is not None:
            l_variable.extend(self.resolution.get_variables())
        if self.phase is not None:
            l_variable.extend(self.phase.get_variables())
        if self.offset.refinement: l_variable.append(self.offset)
        return l_variable


    @property
    def to_cif(self):
        ls_out = []
        ls_out.append("data_{:}\n".format(self.label))
        str_1 = self.params_to_cif
        if str_1 != "":
            ls_out.append(str_1)
        str_2 = self.data_to_cif
        if str_2 != "":
            ls_out.append(str_2)
        str_3 = self.calc_to_cif
        if str_3 != "":
            ls_out.append(str_3)
        return "\n".join(ls_out)

    @property
    def params_to_cif(self):
        ls_out = []
        if self.wavelength is not None:
            ls_out.append("_diffrn_radiation_wavelength {:}".format(self.wavelength))
        if self.field is not None:
            ls_out.append("_diffrn_ambient_field {:}".format(self.field))
        if self.chi2_sum is not None:
            if self.chi2_sum:
                s_line = "True" 
            else: 
                s_line = "False"
            ls_out.append("\n_pd2d_chi2_sum {:}".format(s_line))
        if self.chi2_diff is not None:
            if self.chi2_diff:
                s_line = "True" 
            else: 
                s_line = "False"
            ls_out.append("_pd2d_chi2_diff {:}".format(s_line))
        if self.chi2_up is not None:
            if self.chi2_up:
                s_line = "True" 
            else: 
                s_line = "False"
            ls_out.append("_pd2d_chi2_up {:}".format(s_line))
        if self.chi2_down is not None:
            if self.chi2_down:
                s_line = "True" 
            else: 
                s_line = "False"
            ls_out.append("_pd2d_chi2_down {:}".format(s_line))
        if self.ttheta_min is not None:
            ls_out.append("\n_pd2d_2theta_range_min {:.3f}".format( self.ttheta_min))
        if self.ttheta_max is not None:
            ls_out.append("_pd2d_2theta_range_max {:.3f}".format( self.ttheta_max))
        if self.phi_min is not None:
            ls_out.append("\n_pd2d_phi_range_min {:.3f}".format( self.phi_min))
        if self.ttheta_max is not None:
            ls_out.append("_pd2d_phi_range_max {:.3f}".format( self.phi_max))
        if self.offset is not None:
            ls_out.append("_pd2d_calib_2theta_offset {:}".format(self.offset.print_with_sigma))
        if self.background is not None:
            ls_out.append("\n"+self.background.to_cif)
        if self.exclude is not None:
            ls_out.append("\n"+self.exclude.to_cif)
        if self.asymmetry is not None:
            ls_out.append("\n"+self.asymmetry.to_cif)
        if self.beam_polarization is not None:
            ls_out.append("\n"+self.beam_polarization.to_cif)
        if self.resolution is not None:
            ls_out.append("\n"+self.resolution.to_cif)
        if self.phase is not None:
            ls_out.append("\n"+self.phase.to_cif)
        return "\n".join(ls_out)

    @property
    def data_to_cif(self):
        ls_out = []
        if self.meas is not None:
            ls_out.append("\n"+self.meas.to_cif)
        return "\n".join(ls_out)

    @property
    def calc_to_cif(self):
        ls_out = []
        if self.reflns is not None:
            ls_out.extend(["\n"+_.to_cif for _ in self.reflns])
        if self.peaks is not None:
            ls_out.extend(["\n"+_.to_cif for _ in self.peaks])
        if self.proc is not None:
            ls_out.append("\n"+self.proc.to_cif)
        return "\n".join(ls_out)



    def from_cif(self, string: str):
        cif_data = Data()
        flag = cif_data.take_from_string(string)
        if not flag:
            return False
        self.label = cif_data.name
        cif_values = cif_data.items
        if cif_values is not None:
            if cif_values.is_prefix("_diffrn_radiation_wavelength"):
                self.wavelength = float(cif_values["_diffrn_radiation_wavelength"])
            if cif_values.is_prefix("_diffrn_ambient_field"):
                self.field = float(cif_values["_diffrn_ambient_field"])
            if cif_values.is_prefix("_pd2d_chi2_sum"):
                self.chi2_sum = (cif_values["_pd2d_chi2_sum"].value).strip().lower()=="true"
            if cif_values.is_prefix("_pd2d_chi2_diff"):
                self.chi2_diff = (cif_values["_pd2d_chi2_diff"].value).strip().lower()=="true"
            if cif_values.is_prefix("_pd2d_chi2_sum"):
                self.chi2_up = (cif_values["_pd2d_chi2_up"].value).strip().lower()=="true"
            if cif_values.is_prefix("_pd2d_chi2_down"):
                self.chi2_down = (cif_values["_pd2d_chi2_down"].value).strip().lower()=="true"
            if cif_values.is_prefix("_pd2d_calib_2theta_offset"):
                self.offset = cif_values["_pd2d_calib_2theta_offset"]
            if cif_values.is_prefix("_pd2d_2theta_range_min"):
                self.ttheta_min = float(cif_values["_pd2d_2theta_range_min"])
            if cif_values.is_prefix("_pd2d_2theta_range_max"):
                self.ttheta_max = float(cif_values["_pd2d_2theta_range_max"])
            if cif_values.is_prefix("_pd2d_phi_range_min"):
                self.phi_min = float(cif_values["_pd2d_phi_range_min"])
            if cif_values.is_prefix("_pd2d_phi_range_max"):
                self.phi_max = float(cif_values["_pd2d_phi_range_max"])
            if cif_values.is_prefix("_diffrn_radiation"):
                self.beam_polarization = str(cif_values)
            if cif_values.is_prefix("_pd2d_background_2theta_phi_intensity"): 
                self.background = str(cif_values["_pd2d_background_2theta_phi_intensity"])
            if cif_values.is_prefix("_pd2d_instr_reflex_asymmetry"): 
                self.asymmetry = "\n".join([str(_) for _ in cif_values["_pd2d_instr_reflex_asymmetry"]])
            if cif_values.is_prefix("_pd2d_instr_resolution"): 
                self.resolution = "\n".join([str(_) for _ in cif_values["_pd2d_instr_resolution"]])
            if cif_values.is_prefix("_pd2d_meas_2theta_phi_intensity"): 
                self.meas = "\n".join([str(_) for _ in cif_values["_pd2d_meas_2theta_phi_intensity"]])
        if cif_data.is_prefix("_pd2d_exclude"): self.exclude = str(cif_data["_pd2d_exclude"])
        if cif_data.is_prefix("_pd2d_phase"): self.phase = str(cif_data["_pd2d_phase"])

        return True

    def calc_profile(self, tth, phi, l_crystal, l_peak_in=[], l_refln_in=[], l_dd_in = []):
        """
        calculate intensity for the given diffraction angle
        """
        proc = Pd2dProc() #it is output
        proc.ttheta = tth
        proc.phi = phi
        
        background = self.background
        int_bkgd = background.interpolate_by_points(tth, phi)
        proc.bkg_calc = int_bkgd
        tth_rad = tth*numpy.pi/180.
        phi_rad = phi*numpy.pi/180.
        cos_theta_1d = numpy.cos(0.5*tth_rad)
        sin_phi_1d = numpy.sin(phi_rad)

        wavelength = self.wavelength
        beam_polarization = self.beam_polarization

        p_u = float(beam_polarization.polarization)
        p_d = (2.*float(beam_polarization.efficiency)-1.)*p_u
        

        tth_min = tth.min()
        tth_max = tth.max()+3. 
        if tth_max > 180.:
            tth_max = 180.
        sthovl_min = numpy.sin(0.5*tth_min*numpy.pi/180.)/wavelength
        sthovl_max = numpy.sin(0.5*tth_max*numpy.pi/180.)/wavelength

        res_u_2d = numpy.zeros((tth.shape[0],phi.shape[0]), dtype=float)
        res_d_2d = numpy.zeros((tth.shape[0],phi.shape[0]), dtype=float)


        phase = self.phase
        if len(phase.label) != len(l_peak_in):
            l_peak_in = len(phase.label) * [None] 
        if len(phase.label) != len(l_refln_in):
            l_refln_in = len(phase.label) * [None] 
        if len(phase.label) != len(l_dd_in):
            l_dd_in = len(phase.label) * [None] 
        


        l_peak, l_refln, l_dd_out = [], [], []
        for phase_label, phase_scale, phase_igsize, peak_in, refln_in, dd_in in zip(
            phase.label, phase.scale, phase.igsize, l_peak_in, l_refln_in, l_dd_in):
            dd_out = {}
            for i_crystal, crystal in enumerate(l_crystal):
                if crystal.label == phase_label:
                    ind_cry = i_crystal
                    break
            if ind_cry is None:
                print("Crystal with name '{:}' is not found.".format(
                        phase_label))
                return
            crystal = l_crystal[ind_cry]
            scale = float(phase_scale)
            i_g = float(phase_igsize)


            cell = crystal.cell
            space_group = crystal.space_group
            
            peak = Pd2dPeak()
            if peak_in is not None:
                h, k, l, mult = peak_in.h, peak_in.k, peak_in.l, peak_in.mult
            else:
                h, k, l, mult = cell.calc_hkl(space_group, sthovl_min, sthovl_max)
            peak.h, peak.k, peak.l, peak.mult = h, k, l, mult

                
            cond_1 = not(crystal.is_variable)
            cond_2 = (peak_in is not None) & (refln_in is not None)
            if cond_1 & cond_2:
                f_nucl_sq, f_m_p_sin_sq = peak_in.f_nucl_sq, peak_in.f_m_p_sin_sq
                f_m_p_cos_sq, cross_sin = peak_in.f_m_p_cos_sq, peak_in.cross_sin 
                refln = refln_in
            else:
                f_nucl_sq, f_m_p_sin_sq, f_m_p_cos_sq, cross_sin, refln = self.calc_for_iint(h, k, l, crystal)
            l_refln.append(refln)

            peak.f_nucl_sq, peak.f_m_p_sin_sq = f_nucl_sq, f_m_p_sin_sq
            peak.f_m_p_cos_sq, peak.cross_sin = f_m_p_cos_sq, cross_sin

            cond_1 = dd_in is not None
            cond_2 = ((not(crystal.is_variable)) & 
                      (not(beam_polarization.polarization.refinement)) & 
                      (not(beam_polarization.efficiency.refinement)))
            if cond_1 & cond_2:
                iint_u_3d, iint_d_3d = dd_in["iint_u_3d"], dd_in["iint_d_3d"]
                cos_theta_3d, sin_phi_3d = dd_in["cos_theta_3d"], dd_in["sin_phi_3d"]
            else:                      
                cos_theta_3d, sin_phi_3d, mult_f_n_3d = numpy.meshgrid(cos_theta_1d, sin_phi_1d, mult*f_nucl_sq, indexing="ij")
                mult_f_m_c_3d = numpy.meshgrid(tth_rad, phi_rad, mult*f_m_p_cos_sq, indexing="ij")[2]

                hh_u_s_3d = numpy.meshgrid(tth_rad, phi_rad, mult*(f_m_p_sin_sq+p_u*cross_sin), indexing="ij")[2]
                hh_d_s_3d = numpy.meshgrid(tth_rad, phi_rad, mult*(f_m_p_sin_sq-p_d*cross_sin), indexing="ij")[2]

                c_a_sq_3d = (cos_theta_3d * sin_phi_3d)**2
                s_a_sq_3d = 1.-c_a_sq_3d

                iint_u_3d = (mult_f_n_3d + hh_u_s_3d*s_a_sq_3d + mult_f_m_c_3d*c_a_sq_3d)
                iint_d_3d = (mult_f_n_3d + hh_d_s_3d*s_a_sq_3d + mult_f_m_c_3d*c_a_sq_3d)
            dd_out["cos_theta_3d"] = cos_theta_3d
            dd_out["sin_phi_3d"] = sin_phi_3d
            dd_out["iint_u_3d"] = iint_u_3d
            dd_out["iint_d_3d"] = iint_d_3d

            sthovl_hkl = cell.calc_sthovl(h, k, l)
            
            tth_hkl_rad = 2.*numpy.arcsin(sthovl_hkl*wavelength)
            tth_hkl = tth_hkl_rad*180./numpy.pi
            peak.ttheta = tth_hkl


            cond_1 = dd_in is not None
            cond_2 = ((not(phase_igsize.refinement)) & 
                      (not(self.resolution.is_variable)) & 
                      (not(self.is_variable_offset)))
            if cond_1 & cond_2:
                profile_3d, tth_zs, h_pv = dd_in["profile_3d"], dd_in["tth_zs"], dd_in["h_pv"]
            else:
                profile_3d, tth_zs, h_pv = self.calc_shape_profile(tth, phi, tth_hkl, i_g)
            dd_out["profile_3d"] = profile_3d
            dd_out["tth_zs"] = tth_zs
            dd_out["h_pv"] = h_pv
            peak.width_2theta = h_pv
            
            
            res_u_3d = profile_3d*iint_u_3d 
            res_d_3d = profile_3d*iint_d_3d 

            res_u_2d += scale*res_u_3d.sum(axis=2) 
            res_d_2d += scale*res_d_3d.sum(axis=2) 
            l_peak.append(peak)
            l_dd_out.append(dd_out)
        proc.ttheta_corrected = tth_zs
        proc.up_net = res_u_2d
        proc.down_net = res_d_2d
        proc.up_total = res_u_2d+int_bkgd
        proc.down_total = res_d_2d+int_bkgd
        return proc, l_peak, l_refln, l_dd_out
    #def calc_y_mod(self, l_crystal, d_prof_in={}):
    #    """
    #    calculate model diffraction profiles up and down if observed data is defined
    #    """
    #    observed_data = self._p_observed_data
    #    tth = observed_data.get_val('tth')
    #
    #    wavelength = observed_data.get_val('wavelength')
    #    setup = self._p_setup
    #    setup.set_val(wavelength=wavelength)
    #
    #    field = observed_data.get_val('field')
    #    for calculated_data in self._list_calculated_data:
    #        calculated_data.set_val(field=field)
    #
    #    int_u_mod, int_d_mod, d_exp_prof_out = self.calc_profile(tth, l_crystal, d_prof_in)
    #    return int_u_mod, int_d_mod, d_exp_prof_out

    def calc_chi_sq(self, l_crystal):
        """
        calculate chi square
        """
        meas = self.meas

        tth = meas.ttheta
        phi = meas.phi
        int_u_exp = meas.up
        sint_u_exp = meas.up_sigma
        int_d_exp = meas.down
        sint_d_exp = meas.down_sigma

        if ((self.peaks is not None) & (self.reflns is not None)):
            l_peak_in = self.peaks
            l_refln_in = self.reflns
            l_dd_in = self.__dd
        else:
            l_peak_in, l_refln_in, l_dd_in = [], [], [] 

        cond_tth_in = numpy.ones(tth.size, dtype=bool)
        cond_tth_in = numpy.logical_and(cond_tth_in, tth >= self.ttheta_min)
        cond_tth_in = numpy.logical_and(cond_tth_in, tth <= self.ttheta_max)

        cond_phi_in = numpy.ones(phi.size, dtype=bool)
        cond_phi_in = numpy.logical_and(cond_phi_in, phi >= self.phi_min)
        cond_phi_in = numpy.logical_and(cond_phi_in, phi <= self.phi_max)

        #cond_1_in, cond_2_in = numpy.meshgrid(cond_tth_in, cond_phi_in, indexing="ij")
        #cond_in = numpy.logical_and(cond_1_in, cond_2_in)
        tth_in = tth[cond_tth_in]
        phi_in = phi[cond_phi_in]
        int_u_exp_in = int_u_exp[cond_tth_in, :][:, cond_phi_in]
        sint_u_exp_in = sint_u_exp[cond_tth_in, :][:, cond_phi_in]
        int_d_exp_in = int_d_exp[cond_tth_in, :][:, cond_phi_in]
        sint_d_exp_in = sint_d_exp[cond_tth_in, :][:, cond_phi_in]



        proc, l_peak, l_refln, l_dd_out = self.calc_profile(tth_in, phi_in, l_crystal, l_peak_in, l_refln_in, l_dd_in)
        proc.up = int_u_exp_in
        proc.up_sigma = sint_u_exp_in
        proc.down = int_d_exp_in
        proc.down_sigma = sint_d_exp_in
        self.proc = proc
        self.peaks = l_peak
        self.reflns = l_refln
        self.__dd = l_dd_out

        int_u_mod = proc.up_total
        int_d_mod = proc.down_total

        sint_sum_exp_in = (sint_u_exp_in**2 + sint_d_exp_in**2)**0.5

        chi_sq_u = ((int_u_mod-int_u_exp_in)/sint_u_exp_in)**2
        chi_sq_d = ((int_d_mod-int_d_exp_in)/sint_d_exp_in)**2
        chi_sq_sum = ((int_u_mod+int_d_mod-int_u_exp_in-int_d_exp_in)/sint_sum_exp_in)**2
        chi_sq_dif = ((int_u_mod-int_d_mod-int_u_exp_in+int_d_exp_in)/sint_sum_exp_in)**2

        cond_u = numpy.logical_not(numpy.isnan(chi_sq_u))
        cond_d = numpy.logical_not(numpy.isnan(chi_sq_d))
        cond_sum = numpy.logical_not(numpy.isnan(chi_sq_sum))
        cond_dif = numpy.logical_not(numpy.isnan(chi_sq_dif))

        #exclude region
        exclude = self.exclude
        if exclude is not None:
            l_excl_tth_min = exclude.ttheta_min
            l_excl_tth_max = exclude.ttheta_max
            l_excl_phi_min = exclude.phi_min
            l_excl_phi_max = exclude.phi_max
            for excl_tth_min, excl_tth_max, excl_phi_min, excl_phi_max in zip(l_excl_tth_min, l_excl_tth_max, l_excl_phi_min, l_excl_phi_max):
                cond_1 = numpy.logical_or(tth_in < 1.*excl_tth_min, tth_in > 1.*excl_tth_max)
                cond_2 = numpy.logical_or(phi_in < 1.*excl_phi_min, phi_in > 1.*excl_phi_max)
                cond_11, cond_22 = numpy.meshgrid(cond_1, cond_2, indexing="ij")
                cond_12 = numpy.logical_or(cond_11, cond_22)
                cond_u = numpy.logical_and(cond_u, cond_12)
                cond_d = numpy.logical_and(cond_d, cond_12)
                cond_sum = numpy.logical_and(cond_sum, cond_12)


        chi_sq_u_val = (chi_sq_u[cond_u]).sum()
        n_u = cond_u.sum()
        
        chi_sq_d_val = (chi_sq_d[cond_d]).sum()
        n_d = cond_d.sum()

        chi_sq_sum_val = (chi_sq_sum[cond_sum]).sum()
        n_sum = cond_sum.sum()

        chi_sq_dif_val = (chi_sq_dif[cond_dif]).sum()
        n_dif = cond_dif.sum()

        flag_u = self.chi2_up
        flag_d = self.chi2_down
        flag_sum = self.chi2_sum
        flag_dif = self.chi2_diff
        
        chi_sq_val = (int(flag_u)*chi_sq_u_val + int(flag_d)*chi_sq_d_val + 
                  int(flag_sum)*chi_sq_sum_val + int(flag_dif)*chi_sq_dif_val)
        n = (int(flag_u)*n_u + int(flag_d)*n_d + int(flag_sum)*n_sum + 
             int(flag_dif)*n_dif)
        #d_exp_out = {"chi_sq_val": chi_sq_val, "n": n}
        #d_exp_out.update(d_exp_prof_out)
        return chi_sq_val, n


    def calc_for_iint(self, h, k, l, crystal):
        """
        calculate the integral intensity for h, k, l reflections
        """
        field = float(self.field)
        beam_polarization = self.beam_polarization
        p_u = float(beam_polarization.polarization)
        p_d = (2.*float(beam_polarization.efficiency)-1)*p_u

        refln = crystal.calc_sf(h, k, l)

        f_nucl, sft_11, sft_12, sft_13 = refln.f_nucl, refln.sft_11, refln.sft_12, refln.sft_13
        sft_21, sft_22, sft_23 = refln.sft_21, refln.sft_22, refln.sft_23
        sft_31, sft_32, sft_33 = refln.sft_31, refln.sft_32, refln.sft_33

        cell = crystal.cell
        
        #k_loc = cell.calc_k_loc(h, k, l)
        t_11, t_12, t_13, t_21, t_22, t_23, t_31, t_32, t_33 = cell.calc_m_t(h, k, l)
        
        th_11, th_12, th_13, th_21, th_22, th_23, th_31, th_32, th_33 = calc_mRmCmRT(
                t_11, t_21, t_31, t_12, t_22, t_32, t_13, t_23, t_33,
                sft_11, sft_12, sft_13, sft_21, sft_22, sft_23, sft_31, sft_32, 
                sft_33)


        f_nucl_sq = abs(f_nucl*f_nucl.conjugate())
        f_m_p_sin_sq = (field**2)*abs(0.5*(th_11*th_11.conjugate()+th_22*th_22.conjugate())+th_12*th_12.conjugate())
        f_m_p_cos_sq = (field**2)*abs(th_13*th_13.conjugate()+th_23*th_23.conjugate())
        f_m_p_field = 0.5*field*(th_11+th_22) 
        cross_sin = 2.*(f_nucl.real*f_m_p_field.real+f_nucl.imag*f_m_p_field.imag)
        
        return f_nucl_sq, f_m_p_sin_sq, f_m_p_cos_sq, cross_sin, refln

    def _gauss_pd(self, tth_2d):
        """
        one dimensional gauss powder diffraction
        """
        ag, bg = self.__ag, self.__bg
        val_1 = bg*tth_2d**2
        val_2 = numpy.where(val_1 < 5., numpy.exp(-val_1), 0.)
        self.__gauss_pd = ag*val_2

    def _lor_pd(self, tth_2d):
        """
        one dimensional lorentz powder diffraction
        """
        al, bl = self.__al, self.__bl
        self.__lor_pd = al*1./(1.+bl*tth_2d**2)
    

    def calc_shape_profile(self, tth, phi, tth_hkl, i_g=0.):
        """
        calculate profile in the range ttheta for reflections placed on 
        ttheta_hkl with i_g parameter by default equal to zero
        
        tth, phi, tth_hkl in degrees

        """
        zero_shift = float(self.offset)
        tth_zs = tth-zero_shift

        resolution = self.resolution
        
        h_pv, eta, h_g, h_l, a_g, b_g, a_l, b_l = resolution.calc_resolution(
                                                  tth_hkl, i_g=i_g)


        tth_3d, phi_3d, tth_hkl_3d = numpy.meshgrid(tth_zs, phi, tth_hkl, indexing="ij")


        self.__ag = numpy.meshgrid(tth_zs, phi, a_g, indexing="ij")[2]
        self.__bg = numpy.meshgrid(tth_zs, phi, b_g, indexing="ij")[2]
        self.__al = numpy.meshgrid(tth_zs, phi, a_l, indexing="ij")[2]
        self.__bl = numpy.meshgrid(tth_zs, phi, b_l, indexing="ij")[2]
        eta_3d = numpy.meshgrid(tth_zs, phi, eta, indexing="ij")[2]
        self.__eta = eta_3d 

        self._gauss_pd(tth_3d-tth_hkl_3d)
        self._lor_pd(tth_3d-tth_hkl_3d)
        g_pd2d_3d = self.__gauss_pd 
        l_pd2d_3d = self.__lor_pd
        
        np_shape_3d = eta_3d * l_pd2d_3d + (1.-eta_3d) * g_pd2d_3d

        asymmetry = self.asymmetry
        np_ass_2d = asymmetry.calc_asymmetry(tth_zs, tth_hkl)
        np_ass_3d = np_ass_2d[:, numpy.newaxis,:]*numpy.ones(phi.size, dtype=float)[numpy.newaxis, :, numpy.newaxis]
       
        #Lorentz factor
        tth_rad = tth_zs*numpy.pi/180.
        np_lor_1d = 1./(numpy.sin(tth_rad)*numpy.sin(0.5*tth_rad))
        np_lor_3d = numpy.meshgrid(np_lor_1d, phi, tth_hkl, indexing="ij")[0]
        
        
        profile_3d = np_shape_3d*np_ass_3d*np_lor_3d
        
        return profile_3d, tth_zs, h_pv





__author__ = 'ikibalin'
__version__ = "2019_08_25"
import os
import numpy
from pycifstar import Global

from cryspy.f_common.cl_fitable import Fitable
from cryspy.f_crystal.cl_space_group import SpaceGroup

class Cell(object):
    """
    Data items in the Cell class record details about the
    crystallographic cell parameters and their measurement.

    Description in cif file:

        _cell_length_a                     5.959(1)
        _cell_length_b                     14.956(1)
        _cell_length_c                     19.737(3)
        _cell_angle_alpha                  90
        _cell_angle_beta                   90
        _cell_angle_gamma                  90
    """
    def __init__(self, a = Fitable(1.0), 
                       b = Fitable(1.0), 
                       c = Fitable(1.0), 
                       alpha = Fitable(90.0), 
                       beta = Fitable(90.0), 
                       gamma= Fitable(90.), bravais_lattice = "triclinic"):
        super(Cell, self).__init__()
        self.__cell_length_a = None
        self.__cell_length_b = None
        self.__cell_length_c = None
        self.__cell_angle_alpha = None
        self.__cell_angle_beta = None
        self.__cell_angle_gamma = None
        self.__cell_setting = None

        self.__cos_a = None
        self.__cos_b = None
        self.__cos_g = None
        self.__cos_a_sq = None
        self.__cos_b_sq = None
        self.__cos_g_sq = None
        self.__sin_a = None
        self.__sin_b = None
        self.__sin_g = None
        self.__sin_a_sq = None
        self.__sin_b_sq = None
        self.__sin_g_sq = None
        
        self.__cell_reciprocal_length_a = None
        self.__cell_reciprocal_length_b = None
        self.__cell_reciprocal_length_c = None
        self.__cell_reciprocal_angle_alpha = None
        self.__cell_reciprocal_angle_beta = None
        self.__cell_reciprocal_angle_gamma = None        

        self.__cos_ia = None
        self.__cos_ib = None
        self.__cos_ig = None
        self.__cos_ia_sq = None
        self.__cos_ib_sq = None
        self.__cos_ig_sq = None
        self.__sin_ia = None
        self.__sin_ib = None
        self.__sin_ig = None
        self.__sin_ia_sq = None
        self.__sin_ib_sq = None
        self.__sin_ig_sq = None
        
        self.__cell_volume = None
        self.__cell_ivolume = None
        self.__m_b = None
        self.__m_ib = None
        self.__m_ib_norm = None

        self.a = a # type: Fitable
        self.b = b # type: Fitable
        self.c = c # type: Fitable
        self.alpha = alpha # type: Fitable
        self.beta = beta # type: Fitable
        self.gamma = gamma # type: Fitable
        self.bravais_lattice = bravais_lattice # type: str
        
    @property
    def a(self):
        """
        Unit-cell lengths in angstroms corresponding to the structure
        reported. 
        The permitted range is 0.0 -> infinity

        reference: https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Icell_length_.html
        """
        return self.__cell_length_a
    @a.setter
    def a(self, x):
        if isinstance(x, Fitable):
            x_in = x
        else:
            x_in = Fitable()
            flag = x_in.take_it(x)
        self.__cell_length_a = x_in
        self.apply_constraint()        
    @property
    def b(self):
        """
        Unit-cell lengths in angstroms corresponding to the structure
        reported. 
        The permitted range is 0.0 -> infinity
        
        reference: https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Icell_length_.html
        """
        return self.__cell_length_b
    @b.setter
    def b(self, x):
        if isinstance(x, Fitable):
            x_in = x
        else:
            x_in = Fitable()
            flag = x_in.take_it(x)
        self.__cell_length_b = x_in
        self.apply_constraint()        
    @property
    def c(self):
        """
        Unit-cell lengths in angstroms corresponding to the structure
        reported. 
        The permitted range is 0.0 -> infinity
        
        reference: https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Icell_length_.html
        """
        return self.__cell_length_c
    @c.setter
    def c(self, x):
        if isinstance(x, Fitable):
            x_in = x
        else:
            x_in = Fitable()
            flag = x_in.take_it(x)
        self.__cell_length_c = x_in
        self.apply_constraint()        
    @property
    def alpha(self):
        """
        Unit-cell angles of the reported structure in degrees.
        The permitted range is 0.0 -> 180.0 
        Enumeration default: 90.0
        
        reference: https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Icell_angle_.html
        """
        return self.__cell_angle_alpha
    @alpha.setter
    def alpha(self, x):
        if isinstance(x, Fitable):
            x_in = x
        else:
            x_in = Fitable()
            flag = x_in.take_it(x)
        self.__cell_angle_alpha = x_in
        self.apply_constraint()        
    @property
    def beta(self):
        """
        Unit-cell angles of the reported structure in degrees.
        The permitted range is 0.0 -> 180.0 
        Enumeration default: 90.0
        
        reference: https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Icell_angle_.html
        """
        return self.__cell_angle_beta
    @beta.setter
    def beta(self, x):
        if isinstance(x, Fitable):
            x_in = x
        else:
            x_in = Fitable()
            flag = x_in.take_it(x)
        self.__cell_angle_beta = x_in
        self.apply_constraint()        
    @property
    def gamma(self):
        """
        Unit-cell angles of the reported structure in degrees.
        The permitted range is 0.0 -> 180.0 
        Enumeration default: 90.0
        
        reference: https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Icell_angle_.html
        """
        return self.__cell_angle_gamma
    @gamma.setter
    def gamma(self, x):
        if isinstance(x, Fitable):
            x_in = x
        else:
            x_in = Fitable()
            flag = x_in.take_it(x)
        self.__cell_angle_gamma = x_in
        self.apply_constraint()        
    @property
    def bravais_lattice(self):
        """
        The cell settings for this space-group symmetry.

        The data value must be one of the following:

        triclinic	
        monoclinic	
        orthorhombic	
        tetragonal	
        rhombohedral	
        trigonal	
        hexagonal	
        cubic

        reference: https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Isymmetry_cell_setting.html
        """
        return self.__cell_setting
    @bravais_lattice.setter
    def bravais_lattice(self, x):
        try:
            x_in = str(x).lower()
        except:
            x_in = "triclinic"
        l_bravais_lattice = ["cubic", "hexagonal", "rhombohedral", "trigonal", "tetragonal", 
                             "orthorhombic", "monoclinic", "triclinic"]
        if x_in not in l_bravais_lattice:
            print("Introduced bravais_lattice is not found.")
            print("Please try one of them: {:}.".format(", ".join(l_bravais_lattice)))
            x_in = "Triclinic"
        self.__cell_setting = x_in
        self.apply_constraint()        
    @property
    def volume(self):
        """
        Cell volume V in angstroms cubed.
        
        V = a b c [1 - cos^2^(alpha) - cos^2^(beta) - cos^2^(gamma)
               + 2 cos(alpha) cos(beta) cos(gamma) ] ^1/2^

        The permitted range is 0.0 -> infinity

        reference: https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Icell_volume.html
        """
        return self.__cell_volume
    @property
    def ivolume(self):
        """
        The reciprocal-cell volume V in angstroms cubed.
        (not realized)
        """
        return self.__cell_ivolume
    @property
    def ia(self):
        """
        The reciprocal-cell lengths in inverse angstroms.  These are
        related to the real cell by:

        recip-a = b*c*sin(alpha)/V
        recip-b = c*a*sin(beta)/V
        recip-c = a*b*sin(gamma)/V

        where V is the cell volume.

        Ref: Buerger, M. J. (1942). X-ray Crystallography, p. 360.
           New York: John Wiley & Sons Inc.

        The permitted range is 0.0 -> infinity

        reference: https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Icell_reciprocal_length_.html
        """
        return self.__cell_reciprocal_length_a
    @property
    def ib(self):
        """
        The reciprocal-cell lengths in inverse angstroms.  These are
        related to the real cell by:

        recip-a = b*c*sin(alpha)/V
        recip-b = c*a*sin(beta)/V
        recip-c = a*b*sin(gamma)/V

        where V is the cell volume.

        Ref: Buerger, M. J. (1942). X-ray Crystallography, p. 360.
           New York: John Wiley & Sons Inc.

        The permitted range is 0.0 -> infinity
        
        reference: https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Icell_reciprocal_length_.html
        """
        return self.__cell_reciprocal_length_b
    @property
    def ic(self):
        """
        The reciprocal-cell lengths in inverse angstroms.  These are
        related to the real cell by:

        recip-a = b*c*sin(alpha)/V
        recip-b = c*a*sin(beta)/V
        recip-c = a*b*sin(gamma)/V

        where V is the cell volume.

        Ref: Buerger, M. J. (1942). X-ray Crystallography, p. 360.
           New York: John Wiley & Sons Inc.

        The permitted range is 0.0 -> infinity
        
        reference: https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Icell_reciprocal_length_.html
        """
        return self.__cell_reciprocal_length_c
    @property
    def ialpha(self):
        """
        The angles defining the reciprocal cell in degrees. These
        are related to those in the real cell by:

        cos(recip-alpha)
          = [cos(beta)*cos(gamma) - cos(alpha)]/[sin(beta)*sin(gamma)]

        cos(recip-beta)
          = [cos(gamma)*cos(alpha) - cos(beta)]/[sin(gamma)*sin(alpha)]

        cos(recip-gamma)
         = [cos(alpha)*cos(beta) - cos(gamma)]/[sin(alpha)*sin(beta)]

        Ref: Buerger, M. J. (1942). X-ray Crystallography, p. 360.
           New York: John Wiley & Sons Inc.

        The permitted range is 0.0 -> 180.0 
        Enumeration default: 90.0

        reference: https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Icell_reciprocal_angle_.html
        """
        return self.__cell_reciprocal_angle_alpha
    @property
    def ibeta(self):
        """
        The angles defining the reciprocal cell in degrees. These
        are related to those in the real cell by:

        cos(recip-alpha)
          = [cos(beta)*cos(gamma) - cos(alpha)]/[sin(beta)*sin(gamma)]

        cos(recip-beta)
          = [cos(gamma)*cos(alpha) - cos(beta)]/[sin(gamma)*sin(alpha)]

        cos(recip-gamma)
         = [cos(alpha)*cos(beta) - cos(gamma)]/[sin(alpha)*sin(beta)]

        Ref: Buerger, M. J. (1942). X-ray Crystallography, p. 360.
           New York: John Wiley & Sons Inc.

        The permitted range is 0.0 -> 180.0 
        Enumeration default: 90.0

        reference: https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Icell_reciprocal_angle_.html
        """
        return self.__cell_reciprocal_angle_beta
    @property
    def igamma(self):
        """
        The angles defining the reciprocal cell in degrees. These
        are related to those in the real cell by:

        cos(recip-alpha)
          = [cos(beta)*cos(gamma) - cos(alpha)]/[sin(beta)*sin(gamma)]

        cos(recip-beta)
          = [cos(gamma)*cos(alpha) - cos(beta)]/[sin(gamma)*sin(alpha)]

        cos(recip-gamma)
         = [cos(alpha)*cos(beta) - cos(gamma)]/[sin(alpha)*sin(beta)]

        Ref: Buerger, M. J. (1942). X-ray Crystallography, p. 360.
           New York: John Wiley & Sons Inc.

        The permitted range is 0.0 -> 180.0 
        Enumeration default: 90.0

        reference: https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Icell_reciprocal_angle_.html
        """
        return self.__cell_reciprocal_angle_gamma
    @property
    def m_b(self):
        """
        B matrix
        """
        return self.__m_b
    @property
    def m_ib(self):
        """
        Inversed B matrix
        """
        return self.__m_ib

    @property
    def m_ib_norm(self):
        """
        Normalized inversed B matrix
        """
        return self.__m_ib_norm

    @property
    def cos_a(self):
        """
        cos(alpha)
        """
        return self.__cos_a
    @property
    def cos_b(self):
        """
        cos(beta)
        """
        return self.__cos_b
    @property
    def cos_g(self):
        """
        cos(gamma)
        """
        return self.__cos_g
    @property
    def cos_ib(self):
        """
        cos(beta*)
        """
        return self.__cos_ib
    @property
    def sin_ib(self):
        """
        sin(beta*)
        """
        return self.__sin_ib
    @property
    def cos_ig(self):
        """
        cos(gamma*)
        """
        return self.__cos_ig
    @property
    def sin_ig(self):
        """
        sin(gamma*)
        """
        return self.__sin_ig
    def __repr__(self):
        ls_out = ["Cell:"]
        ls_out.append(str(self))
        return "\n".join(ls_out)
        
    def __str__(self):
        ls_out = [""" a: {:}\n b: {:}\n c: {:}\n alpha: {:}
 beta: {:}\n gamma: {:}\n bravais_lattice: {:}""".format(self.a, self.b, 
                 self.c, self.alpha, self.beta, self.gamma, 
                 self.bravais_lattice)]
        if self.__cell_volume is not None:
            ls_out.append(" volume: {:.3f}".format(self.__cell_volume))

        if self.__m_b is not None:
             ls_out.append(""" B matrix is:\n {:9.5f} {:9.5f} {:9.5f}
 {:9.5f} {:9.5f} {:9.5f}\n {:9.5f} {:9.5f} {:9.5f}""".format(self.__m_b[0, 0],
 self.__m_b[0, 1], self.__m_b[0, 2], self.__m_b[1, 0], self.__m_b[1, 1], 
 self.__m_b[1, 2], self.__m_b[2, 0], self.__m_b[2, 1], self.__m_b[2, 2]))
        if self.__m_ib is not None:
             ls_out.append(""" inversed B matrix is:\n {:9.5f} {:9.5f} {:9.5f}
 {:9.5f} {:9.5f} {:9.5f}\n {:9.5f} {:9.5f} {:9.5f}""".format(self.__m_ib[0, 0],
 self.__m_ib[0, 1], self.__m_ib[0, 2], self.__m_ib[1, 0], self.__m_ib[1, 1], 
 self.__m_ib[1, 2], self.__m_ib[2, 0], self.__m_ib[2, 1], self.__m_ib[2, 2]))
        if self.__m_ib_norm is not None:
             ls_out.append(""" normalized inversed B matrix is:\n {:9.5f} {:9.5f} {:9.5f}
 {:9.5f} {:9.5f} {:9.5f}\n {:9.5f} {:9.5f} {:9.5f}""".format(self.__m_ib_norm[0, 0],
 self.__m_ib_norm[0, 1], self.__m_ib_norm[0, 2], self.__m_ib_norm[1, 0], self.__m_ib_norm[1, 1], 
 self.__m_ib_norm[1, 2], self.__m_ib_norm[2, 0], self.__m_ib_norm[2, 1], self.__m_ib_norm[2, 2]))
        return "\n".join(ls_out)
    
    
    def _constr_bravais_lattice(self):
        bravais_lattice = self.bravais_lattice
        if bravais_lattice == "cubic":
            self.a.constraint_flag = False
            self.__cell_length_b = Fitable(self.a.value, self.a.sigma, False, "_cell_length_b", True)
            self.__cell_length_c = Fitable(self.a.value, self.a.sigma, False, "_cell_length_c", True)
            self.__cell_angle_alpha = Fitable(90., None, False, "_cell_angle_alpha", True)
            self.__cell_angle_beta = Fitable(90., None, False, "_cell_angle_beta", True)
            self.__cell_angle_gamma = Fitable(90., None, False, "_cell_angle_gamma", True)
        elif bravais_lattice == "hexagonal":
            self.a.constraint_flag = False
            self.c.constraint_flag = False
            self.__cell_length_b = Fitable(self.a.value, self.a.sigma, False, "_cell_length_b", True)
            self.__cell_angle_alpha = Fitable(90., None, False, "_cell_angle_alpha", True)
            self.__cell_angle_beta = Fitable(90., None, False, "_cell_angle_beta", True)
            self.__cell_angle_gamma = Fitable(120., None, False, "_cell_angle_gamma", True)
        elif bravais_lattice == "rhombohedral":
            self.a.constraint_flag = False
            self.alpha.constraint_flag = False
            self.__cell_length_b = Fitable(self.a.value, self.a.sigma, False, "_cell_length_b", True)
            self.__cell_length_c = Fitable(self.a.value, self.a.sigma, False, "_cell_length_c", True)
            self.__cell_angle_beta = Fitable(self.alpha.value, None, False, "_cell_angle_beta", True)
            self.__cell_angle_gamma = Fitable(self.alpha.value, None, False, "_cell_angle_gamma", True)
        elif bravais_lattice == "trigonal":
            self.a.constraint_flag = False
            self.alpha.constraint_flag = False
            self.beta.constraint_flag = False
            self.gamma.constraint_flag = False
            self.__cell_length_b = Fitable(self.a.value, self.a.sigma, False, "_cell_length_b", True)
            self.__cell_length_c = Fitable(self.a.value, self.a.sigma, False, "_cell_length_c", True)
        elif bravais_lattice == "tetragonal":
            self.a.constraint_flag = False
            self.c.constraint_flag = False
            self.__cell_length_b = Fitable(self.a.value, self.a.sigma, False, "_cell_length_b", True)
            self.__cell_angle_alpha = Fitable(90., None, False, "_cell_angle_alpha", True)
            self.__cell_angle_beta = Fitable(90., None, False, "_cell_angle_beta", True)
            self.__cell_angle_gamma = Fitable(90., None, False, "_cell_angle_gamma", True)
        elif bravais_lattice == "orthorhombic":
            self.a.constraint_flag = False
            self.b.constraint_flag = False
            self.c.constraint_flag = False
            self.__cell_angle_alpha = Fitable(90., None, False, "_cell_angle_alpha", True)
            self.__cell_angle_beta = Fitable(90., None, False, "_cell_angle_beta", True)
            self.__cell_angle_gamma = Fitable(90., None, False, "_cell_angle_gamma", True)
        elif bravais_lattice == "monoclinic":
            self.a.constraint_flag = False
            self.b.constraint_flag = False
            self.c.constraint_flag = False
            self.beta.constraint_flag = False
            self.__cell_angle_alpha = Fitable(90., None, False, "_cell_angle_alpha", True)
            self.__cell_angle_gamma = Fitable(90., None, False, "_cell_angle_gamma", True)

    def _calc_cos_abc(self):
        rad=numpy.pi/180.
        self.__cos_a = numpy.cos(self.alpha*rad)
        self.__cos_b = numpy.cos(self.beta*rad)
        self.__cos_g = numpy.cos(self.gamma*rad)
        
        self.__sin_a = numpy.sin(self.alpha*rad)
        self.__sin_b = numpy.sin(self.beta*rad)
        self.__sin_g = numpy.sin(self.gamma*rad)
        
        self.__cos_a_sq = self.__cos_a**2
        self.__cos_b_sq = self.__cos_b**2
        self.__cos_g_sq = self.__cos_g**2

        self.__sin_a_sq = 1.-self.__cos_a_sq
        self.__sin_b_sq = 1.-self.__cos_b_sq
        self.__sin_g_sq = 1.-self.__cos_g_sq
        
    def _calc_cos_iabc(self):
        rad=numpy.pi/180.
        self.__cos_ia = numpy.cos(self.__cell_reciprocal_angle_alpha*rad)
        self.__cos_ib = numpy.cos(self.__cell_reciprocal_angle_beta*rad)
        self.__cos_ig = numpy.cos(self.__cell_reciprocal_angle_gamma*rad)
        
        self.__sin_ia = numpy.sin(self.__cell_reciprocal_angle_alpha*rad)
        self.__sin_ib = numpy.sin(self.__cell_reciprocal_angle_beta*rad)
        self.__sin_ig = numpy.sin(self.__cell_reciprocal_angle_gamma*rad)
        
        self.__cos_ia_sq = self.__cos_ia**2
        self.__cos_ib_sq = self.__cos_ib**2
        self.__cos_ig_sq = self.__cos_ig**2

        self.__sin_a_sq = 1.-self.__cos_a_sq
        self.__sin_b_sq = 1.-self.__cos_b_sq
        self.__sin_g_sq = 1.-self.__cos_g_sq

    def _calc_volume(self):
        a = 1.*self.a
        b = 1.*self.b
        c = 1.*self.c
        c_a = self.__cos_a
        c_b = self.__cos_b
        c_g = self.__cos_g
        c_a_sq = self.__cos_a_sq
        c_b_sq = self.__cos_b_sq
        c_g_sq = self.__cos_g_sq
        vol = a*b*c*(1.-c_a_sq-c_b_sq-c_g_sq+2.*c_a*c_b*c_g)**0.5
        self.__cell_volume = vol
        
    
    def _calc_iucp(self):
        """
        calculate inverse unit cell
        """
        irad = 180./numpy.pi

        a = 1.*self.a
        b = 1.*self.b
        c = 1.*self.c
        c_a = self.__cos_a
        c_b = self.__cos_b
        c_g = self.__cos_g
        s_a = self.__sin_a
        s_b = self.__sin_b
        s_g = self.__sin_g
        vol = self.__cell_volume
        
        self.__cell_reciprocal_angle_alpha = numpy.arccos((c_b*c_g-c_a)/(s_b*s_g))*irad
        self.__cell_reciprocal_angle_beta = numpy.arccos((c_g*c_a-c_b)/(s_g*s_a))*irad
        self.__cell_reciprocal_angle_gamma = numpy.arccos((c_a*c_b-c_g)/(s_a*s_b))*irad

        self.__cell_reciprocal_length_a = b*c*s_a/vol
        self.__cell_reciprocal_length_b = c*a*s_b/vol
        self.__cell_reciprocal_length_c = a*b*s_g/vol


    def _calc_m_b(self):
        """
        calculate matrix B 
        """
        c = 1.*self.c

        ia = self.__cell_reciprocal_length_a 
        ib = self.__cell_reciprocal_length_b 
        ic = self.__cell_reciprocal_length_c 
        
        c_a = self.__cos_a
        
        #ic_a = self._p_cos_ia 
        ic_b = self.__cos_ib 
        ic_g = self.__cos_ig 
        #is_a = self._p_sin_ia 
        is_b = self.__sin_ib 
        is_g = self.__sin_ig 
        
        self.__m_b = numpy.array([[ia,  ib*ic_g,  ic*ic_b],
            [0.,  ib*is_g, -ic*is_b*c_a],
            [0.,       0.,  1./c]], dtype = float)

    def _calc_m_ib(self):
        """
        calculate inverse B matrix 
        """
        x1 = self.__m_b[0,0]
        x2 = self.__m_b[1,1]
        x3 = self.__m_b[2,2]
        x4 = self.__m_b[0,1]
        x5 = self.__m_b[0,2]
        x6 = self.__m_b[1,2]
        #B=[[x1,x4,x5],
        #   [0.,x2,x6],
        #   [0.,0.,x3]]
        #it should be checked
        self.__m_ib = numpy.linalg.inv(self.__m_b)
        #y1 = 1./x1
        #y2 = 1./x2
        #y3 = 1./x3
        #y4 = -1.*x4*1./(x1*x2)
        #y6 = -1.*x6*1./(x2*x3)
        #y5 = (x4*x6-x2*x5)*1./(x1*x2*x3)
        #
        #self.__m_ib = numpy.array([[y1,y4,y5],[0.,y2,y6],[0.,0.,y3]], 
        #                           dtype = float)


        m_ib_norm = numpy.copy(self.__m_ib)
        a, b, c = float(self.a), float(self.b), float(self.c)
        m_ib_norm[0, :] /= a
        m_ib_norm[1, :] /= b
        m_ib_norm[2, :] /= c
        self.__m_ib_norm = m_ib_norm
                
    def _show_message(self, s_out: str):
        print("***  Error ***")
        print(s_out)

    
    def calc_sthovl(self, h=None, k=None, l=None, hkl=None, l_hkl=None, f_print=False):
        """
        Calculate sin(theta)/lambda for list of hkl reflections.

        Keyword arguments:
        h, k, l --- Miller indices (example: h=1, k=0, l=2)
        or
        hkl --- a turple of Miller indices (example: hkl=(1, 0, 2))
        or
        l_hkl --- a list of turples of Miller indices (example: l_hkl=[(1, 0, 2), (1, 1, 3)])
        f_print --- a flag to print output information in terminal
        """
        cond_1 = all([h is not None, k is not None, l is not None])
        cond_2 = hkl is not None
        cond_3 = l_hkl is not None
        if cond_1:
            np_h = h # type: numpy.array or float
            np_k = k # type: numpy.array or float
            np_l = l # type: numpy.array or float
        elif cond_2:
            np_h = hkl[0] # type: float
            np_k = hkl[1] # type: float
            np_l = hkl[2] # type: float
        elif cond_3:
            np_h = numpy.array([hh[0] for hh in l_hkl], dtype=float) # type: numpy.array
            np_k = numpy.array([hh[1] for hh in l_hkl], dtype=float) # type: numpy.array
            np_l = numpy.array([hh[2] for hh in l_hkl], dtype=float) # type: numpy.array
        else: 
            self._show_message("Did not found correct input. Expected h, k, l or hkl or l_hkl")
            return
        a = 1.*self.a
        b = 1.*self.b
        c = 1.*self.c
        c_a = self.__cos_a
        c_b = self.__cos_b
        c_g = self.__cos_g
        c_a_sq = self.__cos_a_sq
        c_b_sq = self.__cos_b_sq
        c_g_sq = self.__cos_g_sq
        s_a_sq = self.__sin_a_sq
        s_b_sq = self.__sin_b_sq
        s_g_sq = self.__sin_g_sq

        A=( 1. - c_a_sq - c_b_sq - c_g_sq + 2.*c_a*c_b*c_g)
        B1 = (s_a_sq*(np_h*1./a)**2+s_b_sq*(np_k*1./b)**2+s_g_sq*(np_l*1./c)**2)
        B2 = 2.*(np_k*np_l*c_a)/(b*c)+2.*(np_h*np_l*c_b)/(a*c)+2.*(np_h*np_k*c_g)/(a*b)
        #it should be checked, I am not sure
        B = B1-B2
        inv_d = (B*1./A)**0.5
        res = 0.5*inv_d
        if f_print:
            ls_out = ["    h     k     l     sthovl"]
            try:
                _ = (hh for hh in res)
                f_iter = True # type: bool
            except TypeError:
                f_iter = False # type: bool
            if f_iter:
                for hh_1, hh_2, hh_3, hh_4 in zip(np_h, np_k, np_l, res):
                    ls_out.append("{:5.1f} {:5.1f} {:5.1f} {:10.5f}".format(hh_1, hh_2, hh_3, hh_4))
            else:
                ls_out.append("{:5.1f} {:5.1f} {:5.1f} {:10.5f}".format(np_h, np_k, np_l, res))
            print("\n".join(ls_out))
        return res

    def calc_k_loc(self, h, k, l):
        """
        Calculate unity scattering vector.

        Keyword arguments:
        h, k, l -- Miller indices
        """
        m_b = self.__m_b
        k_x = m_b[0, 0]*h + m_b[0, 1]*k +m_b[0, 2]*l
        k_y = m_b[1, 0]*h + m_b[1, 1]*k +m_b[1, 2]*l
        k_z = m_b[2, 0]*h + m_b[2, 1]*k +m_b[2, 2]*l
        
        k_norm = (k_x**2 + k_y**2 + k_z**2)**0.5
        if not((type(h) is float)|(type(h) is int)|(type(h) is numpy.float64)):
            k_norm[k_norm == 0.] = 1.
        elif k_norm == 0.:
            k_norm = 1.
        
        k_x = k_x/k_norm
        k_y = k_y/k_norm
        k_z = k_z/k_norm
        
        return k_x, k_y, k_z
        
    def calc_m_t(self, h, k, l):
        """
        define rotation matrix to have new z axis along kloc
        Rotation matrix is defined by Euler angles
        """
        m_b = self.__m_b
        k_x = m_b[0, 0]*h + m_b[0, 1]*k +m_b[0, 2]*l
        k_y = m_b[1, 0]*h + m_b[1, 1]*k +m_b[1, 2]*l
        k_z = m_b[2, 0]*h + m_b[2, 1]*k +m_b[2, 2]*l
        
        k_norm = (k_x**2 + k_y**2 + k_z**2)**0.5
        k_norm[k_norm == 0.] = 1.
        
        k_x = k_x/k_norm
        k_y = k_y/k_norm
        k_z = k_z/k_norm
        

        al = numpy.zeros(k_x.shape, dtype=float)
        
        be = numpy.arccos(k_z)
        sb = numpy.sin(be)
        flag = (sb != 0.)
        
        sa1 = k_x[flag]*1./sb[flag]
        ca2 = -1*k_y[flag]*1./sb[flag]
        sa1[sa1>1] = 1.
        sa1[sa1<-1] = -1.
            
        ca2[ca2>1] = 1.
        ca2[ca2<-1] = -1.

        al1 = numpy.arcsin(sa1)
        al2 = numpy.arccos(ca2)
        
        al_sh = numpy.copy(al1)
        al_sh[sa1 > 0.] = al2[sa1 > 0.]
        al_sh[sa1 <= 0.] = 2.*numpy.pi-al2[sa1 <= 0.]
        al_sh[numpy.abs(al2-al1)<0.00001] = al1[numpy.abs(al2-al1)<0.00001]

        al[flag] = al_sh
            
        ga=0.
        ca, cb, cg = numpy.cos(al), numpy.cos(be), numpy.cos(ga)
        sa, sb, sg = numpy.sin(al), numpy.sin(be), numpy.sin(ga)
        t_11, t_12, t_13 = ca*cg-sa*cb*sg, -ca*sg-sa*cb*cg,  sa*sb
        t_21, t_22, t_23 = sa*cg+ca*cb*sg, -sa*sg+ca*cb*cg, -ca*sb
        t_31, t_32, t_33 =          sb*sg,           sb*cg,     cb
        
        flag = (((sa*sb-k_x)**2+(-ca*sb-k_y)**2+(cb-k_z)**2)>0.0001)
        if any(flag):
            print("Mistake with k_loc")
            print("Program is stopped")
            quit()
        return t_11, t_12, t_13, t_21, t_22, t_23, t_31, t_32, t_33 

    @property
    def is_variable(self):
        """
        Output: True if there is any refined parameter
        """
        res = any([self.a.refinement,
                   self.b.refinement,
                   self.c.refinement,
                   self.alpha.refinement,
                   self.beta.refinement,
                   self.gamma.refinement])
        return res
        
    def get_variables(self):
        """
        Output: the list of the refined parameters
        """
        l_variable = []
        if self.a.refinement:
            l_variable.append(self.a)
        if self.b.refinement:
            l_variable.append(self.b)
        if self.c.refinement:
            l_variable.append(self.c)
        if self.alpha.refinement:
            l_variable.append(self.alpha)
        if self.beta.refinement:
            l_variable.append(self.beta)
        if self.gamma.refinement:
            l_variable.append(self.gamma)
        return l_variable
    
    def apply_constraint(self):
        if self.is_defined:
            self._constr_bravais_lattice()
            self._calc_cos_abc()
            self._calc_volume()
            self._calc_iucp()
            self._calc_cos_iabc()
            self._calc_m_b()
            self._calc_m_ib()

    @property
    def is_defined(self):
        """
        Output: True if all started parameters are given
        """

        cond = any([self.a is None, self.b is None, self.c is None,
                    self.alpha is None, self.beta is None, self.gamma is None,
                    self.bravais_lattice is None])
        return not(cond)

    def calc_hkl(self, space_group, sthovl_min, sthovl_max):
        """
        a list of reflections hkl for cell in the range sthovl_min, sthovl_max
        taking into account the space group
        """
        if not(self.is_defined):
            print("Object 'Cell' is not fully defined for calculations")
            return None
        lhkl,lmult=[],[]
        l_hklres=[]

        hmax = int(2.*self.a*sthovl_max)
        kmax = int(2.*self.b*sthovl_max)
        lmax = int(2.*self.c*sthovl_max)
        hmin, kmin, lmin = -1*hmax, -1*kmax, -1*lmax

        hmin=0
        
        l_orig = space_group.orig
        l_symm = space_group.el_symm
        for h in range(hmin,hmax+1,1):
            for k in range(kmin,kmax+1,1):
                for l in range(lmin,lmax+1,1):
                    flag=(abs(sum([numpy.exp(2.*numpy.pi*1j*(orig[0]*h+orig[1]*k+orig[2]*l)) for orig in l_orig]))>0.00001)
                    #flag=True
                    if (flag):
                        lhkls=[(h*symm[1]+k*symm[5]+l*symm[9], h*symm[2]+k*symm[6]+l*symm[10], h*symm[3]+k*symm[7]+l*symm[11]) for symm in l_symm]
                        lhkls.extend([(-hkl[0],-hkl[1],-hkl[2]) for hkl in lhkls])
                        lhkls.sort(key=lambda x:10000*x[0]+100*x[1]+x[2])
                        if (not(lhkls[-1] in lhkl)):
                            lhkl.append(lhkls[-1])
                            lmult.append(len(set(lhkls)))
                            
        l_hklsthovl=[(hkl, self.calc_sthovl(hkl[0], hkl[1], hkl[2]), mult) for hkl, mult in zip(lhkl, lmult)]
        l_hklsthovl.sort(key=lambda x: x[1])
        l_hklres = [hklsthovl[0] for hklsthovl in l_hklsthovl if ((hklsthovl[1]>sthovl_min) & (hklsthovl[1]<sthovl_max))]
        l_multres = [hklsthovl[2] for hklsthovl in l_hklsthovl if ((hklsthovl[1]>sthovl_min) & (hklsthovl[1]<sthovl_max))]

        h = numpy.array([hh[0] for hh in l_hklres], dtype=int)
        k = numpy.array([hh[1] for hh in l_hklres], dtype=int)
        l = numpy.array([hh[2] for hh in l_hklres], dtype=int)
        mult = numpy.array(l_multres, dtype=int)
        return h, k, l, mult


    def calc_hkl_in_range(self, sthovl_min, sthovl_max):
        """
        give a list of reflections hkl for cell in the range sthovl_min, sthovl_max
        """
        if not(self.is_defined):
            print("Object 'Cell' is not fully defined for calculations")
            return None
        h_max = int(2.*self.a*sthovl_max)
        k_max = int(2.*self.b*sthovl_max)
        l_max = int(2.*self.c*sthovl_max)
        h_min, k_min, l_min = -1*h_max, -1*k_max, -1*l_max

        np_h = numpy.array(range(h_min, h_max+1, 1), dtype=int)
        np_k = numpy.array(range(k_min, k_max+1, 1), dtype=int)
        np_l = numpy.array(range(l_min, l_max+1, 1), dtype=int)
        h_3d, k_3d, l_3d = numpy.meshgrid(np_h, np_k, np_l, indexing="ij")
        
        sthovl_3d = self.calc_sthovl(h_3d, k_3d, l_3d)
        flag_1 = sthovl_3d >= sthovl_min
        flag_2 = sthovl_3d <= sthovl_max
        flag_12 = numpy.logical_and(flag_1, flag_2)
        
        h = h_3d[flag_12]
        k = k_3d[flag_12]
        l = l_3d[flag_12]
        mult = numpy.ones(h.size, dtype=int)
        sthovl = sthovl_3d[flag_12]
        arg_sort = numpy.argsort(sthovl)
        return h[arg_sort], k[arg_sort], l[arg_sort], mult[arg_sort] 

    @property
    def to_cif(self):
        ls_out = ["_cell_length_a {:}".format(self.a.print_with_sigma)]
        ls_out.append("_cell_length_b {:}".format(self.b.print_with_sigma))
        ls_out.append("_cell_length_c {:}".format(self.c.print_with_sigma))
        ls_out.append("_cell_angle_alpha {:}".format(self.alpha.print_with_sigma))
        ls_out.append("_cell_angle_beta {:}".format(self.beta.print_with_sigma))
        ls_out.append("_cell_angle_gamma {:}".format(self.gamma.print_with_sigma))
        return "\n".join(ls_out)
    
    def from_cif(self, string: str):
        cif_global = Global()
        flag = cif_global.take_from_string(string)
        if not flag:
            return False
        flag = False
        if cif_global.is_value("cell_length_a"):
            self.a = cif_global["cell_length_a"] # CIFvalue
        if cif_global.is_value("cell_length_b"):
            self.b = cif_global["cell_length_b"] # CIFvalue
        if cif_global.is_value("cell_length_c"):
            self.c = cif_global["cell_length_c"] # CIFvalue
        if cif_global.is_value("cell_angle_alpha"):
            self.alpha = cif_global["cell_angle_alpha"] # CIFvalue
        if cif_global.is_value("cell_angle_beta"):
            self.beta = cif_global["cell_angle_beta"] # CIFvalue
        if cif_global.is_value("cell_angle_gamma"):
            self.gamma = cif_global["cell_angle_gamma"] # CIFvalue
        if (cif_global.is_value("_space_group_name_H-M_alt") & cif_global.is_value("_space_group_it_coordinate_system_code")):
            space_groupe = SpaceGroup(spgr_given_name=cif_global["_space_group_name_H-M_alt"].value, 
                                      spgr_choice=cif_global["_space_group_it_coordinate_system_code"].value)
            self.bravais_lattice =space_groupe.bravais_lattice

        return True




"""
define class PdMeas which describes the 1d powder diffraction experiment
"""
__author__ = 'ikibalin'
__version__ = "2019_09_06"
import os
import numpy


from pycifstar import Global


class PdMeas(object):
    """
    This section contains the measured diffractogram and information
    about the conditions used for the measurement of the diffraction 
    data set, prior to processing and application of correction
    terms. While additional information may be added to the CIF
    as data are processed and transported between laboratories
    (possibly with the addition of a new _pd_block_id entry), the
    information in this section of the CIF will rarely be changed
    once data collection is complete.

    Where possible, measurements in this section should have no
    post-collection processing applied (normalizations, corrections,
    smoothing, zero-offset corrections etc.). Such corrected
    measurements should be recorded in the _pd_proc_ section.

    Data sets that are measured as counts, where a standard
    uncertainty can be considered equivalent to the standard
    deviation and where the standard deviation can be estimated
    as the square root of the number of counts recorded, should
    use the _pd_meas_counts_ fields. All other intensity values
    should be recorded using _pd_meas_intensity_.

    Example:

    loop_
    _pd_meas_angle_2theta
    _pd_meas_intensity_up
    _pd_meas_intensity_up_sigma
    _pd_meas_intensity_down
    _pd_meas_intensity_down_sigma
     4.00   465.80000   128.97000   301.88000   129.30000
     4.20   323.78000   118.22000   206.06000   120.00000
    
    reference: https://www.iucr.org/__data/iucr/cifdic_html/1/cif_pd.dic/Cpd_meas.html
    """
    def __init__(self, ttheta=[], up=[], down=[], up_sigma=[], down_sigma=[]
                 ):
        super(PdMeas, self).__init__()
        self.__pd_meas_angle_2theta = None
        self.__pd_meas_intensity_up = None
        self.__pd_meas_intensity_up_sigma = None
        self.__pd_meas_intensity_down = None
        self.__pd_meas_intensity_down_sigma = None

        self.ttheta = ttheta
        self.up = up
        self.up_sigma = up_sigma
        self.down = down
        self.down_sigma = down_sigma

    @property
    def ttheta(self):
        return self.__pd_meas_angle_2theta
    @ttheta.setter
    def ttheta(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = float(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=float)
        self.__pd_meas_angle_2theta = np_x_in


    @property
    def up(self):
        return self.__pd_meas_intensity_up
    @up.setter
    def up(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = float(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=float)
        self.__pd_meas_intensity_up = np_x_in

    @property
    def up_sigma(self):
        return self.__pd_meas_intensity_up_sigma
    @up_sigma.setter
    def up_sigma(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = float(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=float)
        self.__pd_meas_intensity_up_sigma = np_x_in

    @property
    def down(self):
        return self.__pd_meas_intensity_down
    @down.setter
    def down(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = float(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=float)
        self.__pd_meas_intensity_down = np_x_in

    @property
    def down_sigma(self):
        return self.__pd_meas_intensity_down_sigma
    @down_sigma.setter
    def down_sigma(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = float(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=float)
        self.__pd_meas_intensity_down_sigma = np_x_in


            
    def __repr__(self):
        ls_out = ["PdMeas:"]
        ls_out.append(" ttheta           up     up_sigma         down   down_sigma")
        for _1, _2, _3, _4, _5 in zip(self.ttheta, self.up, self.up_sigma, self.down, self.down_sigma):
            ls_out.append("{:7.2f} {:12.5f} {:12.5f} {:12.5f} {:12.5f}".format(_1, _2, _3, _4, _5))
        return "\n".join(ls_out)

    @property
    def to_cif(self):
        ls_out = []
        if self.is_defined:
            ls_out.append("loop_")
            ls_out.append("_pd_meas_angle_2theta")
            ls_out.append("_pd_meas_intensity_up")
            ls_out.append("_pd_meas_intensity_up_sigma")
            ls_out.append("_pd_meas_intensity_down")
            ls_out.append("_pd_meas_intensity_down_sigma")
            for _1, _2, _3, _4, _5 in zip(self.ttheta, self.up, self.up_sigma, self.down, self.down_sigma):
                ls_out.append("{:} {:} {:} {:} {:}".format(_1, _2, _3, _4, _5))
        return "\n".join(ls_out)

    def from_cif(self, string: str):
        cif_global = Global()
        flag = cif_global.take_from_string(string)
        if not flag:
            return False
        flag = False
        flag = cif_global.is_prefix("_pd_meas")
        if flag:
            cif_loop = cif_global["_pd_meas"]
            l_name = cif_loop.names
            if "_pd_meas_angle_2theta" in l_name:
                self.ttheta = [float(_1) for _1 in cif_loop["_pd_meas_angle_2theta"]]
            if "_pd_meas_intensity_up" in l_name:
                self.up = [float(_1) for _1 in cif_loop["_pd_meas_intensity_up"]]
            if "_pd_meas_intensity_up_sigma" in l_name:
                self.up_sigma = [float(_1) for _1 in cif_loop["_pd_meas_intensity_up_sigma"]]
            if "_pd_meas_intensity_down" in l_name:
                self.down = [float(_1) for _1 in cif_loop["_pd_meas_intensity_down"]]
            if "_pd_meas_intensity_down_sigma" in l_name:
                self.down_sigma = [float(_1) for _1 in cif_loop["_pd_meas_intensity_down_sigma"]]
        else:
            self.ttheta, self.up, self.up_sigma, self.down, self.down_sigma = [], [], [], [], []
        return True

    @property
    def is_defined(self):
        cond = all([self.ttheta is not None, self.up is not None, self.up_sigma is not None, self.down is not None, self.down_sigma is not None])
        return cond

    @property
    def is_variable(self):
        return False
    
    def get_variables(self):
        return []

    def _show_message(self, s_out: str):
        print("***  Error ***")
        print(s_out)



"""
define classes to describe observed data
"""

__author__ = 'ikibalin'
__version__ = "2019_04_16"
import os
import numpy
import sys


class ObservedDataSingleDomain(dict):
    """
    Containt the experimental data
    """
    def __init__(self, h=None, k=None, l=None, flip_ratio=None, 
                 sflip_ratio=None, wave_length=None, field=None, 
                 orientation=None, file_dir=".", file_name=None):
        super(ObservedDataSingleDomain, self).__init__()
        self._p_h = None#2d array (hkl, domains)
        self._p_k = None#2d array (hkl, domains)
        self._p_l = None#2d array (hkl, domains)
        self._p_flip_ratio = None
        self._p_sflip_ratio = None
        
        self._p_file_dir = None
        self._p_file_name = None
        
        self._p_wave_length = None
        self._p_field = None
        self._p_orientation = None#3d array (d1, d2, domains)
        
        self._refresh(h, k, l, flip_ratio, sflip_ratio, wave_length, field, 
                 orientation, file_dir, file_name)

    def __repr__(self):
        ls_out = """ObservedDataSingleDomain:\n file_dir: {:}
 file_name: {:}""".format(self._p_file_dir, self._p_file_name)
        if self._p_h is not None:
            ls_out += "\n h range: {:} --- {:}".format(
                    self._p_h.min(), self._p_h.max())
        if self._p_k is not None:
            ls_out += "\n k range: {:} --- {:}".format(
                    self._p_k.min(), self._p_k.max())
        if self._p_l is not None:
            ls_out += "\n l range: {:} --- {:}".format(
                    self._p_l.min(), self._p_l.max())
            ls_out += "\n number of reflections: {:}".format(
                    self._p_l.size)
        ls_out += "\n field: {:}".format(self._p_field)
        ls_out += "\n wave_length: {:}".format(self._p_wave_length)
        ls_out += "\n orientation: {:}".format(self._p_orientation)
        return ls_out

    def _refresh(self, h, k, l, flip_ratio, sflip_ratio, wave_length, field, 
                 orientation, file_dir, file_name):
        f_read_data = False
        if h is not None:
            self._p_h = h
        if k is not None:
            self._p_k = k
        if l is not None:
            self._p_l = l
        if flip_ratio is not None:
            self._p_flip_ratio = flip_ratio
        if sflip_ratio is not None:
            self._p_sflip_ratio = sflip_ratio
        if wave_length is not None:
            self._p_wave_length = wave_length
        if field is not None:
            self._p_field = field
        if orientation is not None:
            self._p_orientation = orientation
        if file_dir is not None:
            f_read_data = True
            self._p_file_dir = file_dir
        if file_name is not None:
            f_read_data = True
            if os.path.dirname(file_name) != "":
                self._p_file_name = os.path.basename(file_name)
                self._p_file_dir = os.path.dirname(file_name)
            else:
                self._p_file_name = file_name
        if (f_read_data & (self._p_file_name is not None)):
            if os.path.isfile(os.path.join(self._p_file_dir, self._p_file_name)):
                self.read_data()
            
    def set_val(self,  h=None, k=None, l=None, flip_ratio=None, 
                 sflip_ratio=None, wave_length=None, field=None, 
                 orientation=None, file_dir=None, file_name=None):
        self._refresh(h, k, l, flip_ratio, sflip_ratio, wave_length, field, 
                 orientation, file_dir, file_name)
        
    def get_val(self, label):
        lab = "_p_"+label
        
        if lab in self.__dict__.keys():
            val = self.__dict__[lab]
            if isinstance(val, type(None)):
                self.set_val()
                val = self.__dict__[lab]
        else:
            print("The value '{:}' is not found".format(lab))
            val = None
        return val

    def list_vals(self):
        """
        give a list of parameters with small descripition
        """
        lsout = """
Parameters:
 h, k, l are the Miller indices
 flip_ratio is the flipping ratio
 sflip_ratio is the errorbar of flipping ratio
 wave_length is the neutron wave length in the Angtrems
 field is the magnetic field along z axis 
 orientation is the tranbsformation matrix from local coordinate system to global (matrix U)
 file_dir is directory of file with measured data
 file_name is basename of file with measured data
 
        """
        print(lsout)
    
    def read_data(self):
        """
        read file from file
        """
        finp = os.path.join(self._p_file_dir, self._p_file_name)
        
        ddata = {}
        if not(os.path.isfile(finp)):
            return
        fid = open(finp,'r')
        lcontentH = fid.readlines()
        fid.close()
        lparam = [line[1:].strip() for line in lcontentH if line.startswith('#')]
        if (len(lparam) > 1):
            for line in lparam[:-1]:
                lhelp = line.strip().split()
                if (len(lhelp) > 2):
                    ddata[lhelp[0]] = [float(hh) for hh in lhelp[1:]]
                elif (len(lhelp) == 2):
                    ddata[lhelp[0]] = float(lhelp[1])
                else:
                    print("Mistake in experimental file '{:}' in line:\n {:}".format(finp, line))
                    print("The program is stopped.")
                    quit()
        lnames = lparam[-1].split()
        for name in lnames:
            ddata[name] = []
        lcontent = [line for line in lcontentH if line[0]!='#']
        for line in lcontent:
            for name, val in zip(lnames, line.strip().split()):
                ddata[name].append(val)
                
        field = ddata["field"]
        wave_length = ddata["wave_length"]
        
        l_key = list(ddata.keys())
        n_domain = max([int(hh[12:]) for hh in l_key if hh.startswith("orientation")])
        
        np_h = numpy.array([])
        np_k = numpy.array([])
        np_l = numpy.array([])
        np_o = numpy.array([])
        for i_domain in range(1, n_domain+1):
            l_o = ddata["orientation_{:}".format(i_domain)]
            orientation = numpy.array([[l_o[0], l_o[1], l_o[2]], 
                                   [l_o[3], l_o[4], l_o[5]], 
                                   [l_o[6], l_o[7], l_o[8]]], dtype=float)

            h = numpy.array([int(hh) if hh!="None" else numpy.nan for hh in ddata["h_{:}".format(i_domain)]])
            k = numpy.array([int(hh) if hh!="None" else numpy.nan for hh in ddata["k_{:}".format(i_domain)]])
            l = numpy.array([int(hh) if hh!="None" else numpy.nan for hh in ddata["l_{:}".format(i_domain)]])
            if np_h.shape == (0,):
                np_h = h[:, numpy.newaxis]
                np_k = k[:, numpy.newaxis]
                np_l = l[:, numpy.newaxis]
                np_o = orientation[:,:, numpy.newaxis]
            else:
                np_h = numpy.concatenate((np_h, h[:, numpy.newaxis]), axis=1)
                np_k = numpy.concatenate((np_k, k[:, numpy.newaxis]), axis=1)
                np_l = numpy.concatenate((np_l, l[:, numpy.newaxis]), axis=1)
                np_o = numpy.concatenate((np_o, orientation[:,:, numpy.newaxis]), axis=2)
            
        flip_ratio = numpy.array(ddata["FR"], dtype=float)
        sflip_ratio = numpy.array(ddata["sFR"], dtype=float)
        self.set_val(h=np_h, k=np_k, l=np_l, 
                     flip_ratio=flip_ratio, sflip_ratio=sflip_ratio, 
                     wave_length=wave_length, field=field, 
                     orientation=np_o)

    def create_input_file(self, f_inp=None):
        if f_inp is not None:
            f_dir= os.path.dirname(f_inp)
            f_name= os.path.bathename(f_inp)
            self.set_val(file_dir=f_dir, file_name=f_name)
        f_dir = self._p_file_dir
        f_name = self._p_file_name
        f_full = os.path.join(f_dir, f_name)
        
        s_out = """#wave_length 1.40 
#field  1.000
#orientation_1 0.6468462   -0.6860854   0.3300297 0.2141139   -0.2557555   -0.9343334 0.7319312   0.6810804    -0.0183183 
# h_1  k_1  l_1        FR       sFR
    0    0    8   0.64545   0.01329 """
    
        fid = open(f_full, "w")
        fid.write(s_out)
        fid.close()
        
        
if (__name__ == "__main__"):
  pass






"""
define class Pd2dInstrResolution to describe resolution of 2d powder diffractometer
"""
__author__ = 'ikibalin'
__version__ = "2019_09_10"
import os
import numpy


from pycifstar import Data
from cryspy.f_common.cl_fitable import Fitable


#Description of setup class

class Pd2dInstrResolution(dict):
    """
    Pd2dInstrResolution describes resolution of 2d powder diffractometer

    Example:

    _pd2d_instr_resolution_u 16.9776            
    _pd2d_instr_resolution_v -2.83572490939
    _pd2d_instr_resolution_w 0.576309204074
    _pd2d_instr_resolution_x 0.0
    _pd2d_instr_resolution_y 0.0
    """
    def __init__(self, u = Fitable(0.), v = Fitable(0.), w = Fitable(0.01), 
                       x = Fitable(0.), y = Fitable(0.)):
        super(Pd2dInstrResolution, self).__init__()
        self.__pd2d_instr_resolution_u = None
        self.__pd2d_instr_resolution_v = None
        self.__pd2d_instr_resolution_w = None
        self.__pd2d_instr_resolution_x = None
        self.__pd2d_instr_resolution_y = None

        self.__tan_th = None
        self.__tan_th_sq = None
        self.__cos_th = None
        self.__hg = None
        self.__hl = None
        self.__hpv = None
        self.__eta = None
        self.__ag = None
        self.__bg = None
        self.__al = None
        self.__bl = None

        self.u = u
        self.v = v
        self.w = w
        self.x = x
        self.y = y

    @property
    def u(self):
        return self.__pd2d_instr_resolution_u
    @u.setter
    def u(self, x):
        if isinstance(x, Fitable):
            x_in = x
        else:
            x_in = Fitable()
            flag = x_in.take_it(x)
        self.__pd2d_instr_resolution_u = x_in

    @property
    def v(self):
        return self.__pd2d_instr_resolution_v
    @v.setter
    def v(self, x):
        if isinstance(x, Fitable):
            x_in = x
        else:
            x_in = Fitable()
            flag = x_in.take_it(x)
        self.__pd2d_instr_resolution_v = x_in

    @property
    def w(self):
        return self.__pd2d_instr_resolution_w
    @w.setter
    def w(self, x):
        if isinstance(x, Fitable):
            x_in = x
        else:
            x_in = Fitable()
            flag = x_in.take_it(x)
        self.__pd2d_instr_resolution_w = x_in

    @property
    def x(self):
        return self.__pd2d_instr_resolution_x
    @x.setter
    def x(self, x):
        if isinstance(x, Fitable):
            x_in = x
        else:
            x_in = Fitable()
            flag = x_in.take_it(x)
        self.__pd2d_instr_resolution_x = x_in

    @property
    def y(self):
        return self.__pd2d_instr_resolution_y
    @y.setter
    def y(self, x):
        if isinstance(x, Fitable):
            x_in = x
        else:
            x_in = Fitable()
            flag = x_in.take_it(x)
        self.__pd2d_instr_resolution_y = x_in

    def __repr__(self):
        ls_out = ["Pd2dInstrResolution:"]
        ls_out.append(" u: {:}".format(self.u.print_with_sigma))
        ls_out.append(" v: {:}".format(self.v.print_with_sigma))
        ls_out.append(" w: {:}".format(self.w.print_with_sigma))
        ls_out.append(" x: {:}".format(self.x.print_with_sigma))
        ls_out.append(" y: {:}".format(self.y.print_with_sigma))
        return "\n".join(ls_out)

    @property
    def is_variable(self):
        """
        Output: True if there is any refined parameter
        """
        res = any([self.u.refinement, 
                   self.v.refinement,
                   self.w.refinement,
                   self.x.refinement,
                   self.y.refinement])
        return res        
    
    def get_variables(self):
        """
        Output: the list of the refined parameters
        """
        l_variable = []
        if self.u.refinement:
            l_variable.append(self.u)
        if self.v.refinement:
            l_variable.append(self.v)
        if self.w.refinement:
            l_variable.append(self.w)
        if self.x.refinement:
            l_variable.append(self.x)
        if self.y.refinement:
            l_variable.append(self.y)
        return l_variable

    def _show_message(self, s_out: str):
        print("***  Error ***")
        print(s_out)
        
    @property
    def to_cif(self):
        ls_out = []
        ls_out.append("_pd2d_instr_resolution_u {:}".format(self.u.print_with_sigma))
        ls_out.append("_pd2d_instr_resolution_v {:}".format(self.v.print_with_sigma))
        ls_out.append("_pd2d_instr_resolution_w {:}".format(self.w.print_with_sigma))
        ls_out.append("_pd2d_instr_resolution_x {:}".format(self.x.print_with_sigma))
        ls_out.append("_pd2d_instr_resolution_y {:}".format(self.y.print_with_sigma))
        return "\n".join(ls_out)


    def from_cif(self, string: str):
        cif_data = Data()
        flag = cif_data.take_from_string(string)
        if not flag:
            return False
        flag = False
        if cif_data.is_value("_pd2d_instr_resolution_u"):
            self.u = cif_data["_pd2d_instr_resolution_u"] # CIFvalue
        if cif_data.is_value("_pd2d_instr_resolution_v"):
            self.v = cif_data["_pd2d_instr_resolution_v"] # CIFvalue
        if cif_data.is_value("_pd2d_instr_resolution_w"):
            self.w = cif_data["_pd2d_instr_resolution_w"] # CIFvalue
        if cif_data.is_value("_pd2d_instr_resolution_x"):
            self.x = cif_data["_pd2d_instr_resolution_x"] # CIFvalue
        if cif_data.is_value("_pd2d_instr_resolution_y"):
            self.y = cif_data["_pd2d_instr_resolution_y"] # CIFvalue
        return True


    
    def _calc_tancos(self, th_hkl):
        """
        tth_hkl in radianas
        calculate tangenth (theta)
        """
        self.__t_th = numpy.tan(th_hkl)
        self.__t_th_sq = self.__t_th**2
        res = numpy.cos(th_hkl)

        self.__c_th = res
        self.__ic_th = 1./res
        
    def _calc_hg(self, i_g = 0.):
        """
        ttheta in radians, could be array
        gauss size
        """
        u, v, w = float(self.u), float(self.v), float(self.w)
        res_sq = (u*self.__t_th_sq + v*self.__t_th + w + 
                  i_g*self.__ic_th**2)
        self.__hg = numpy.sqrt(res_sq)
        
    def _calc_hl(self):
        """
        ttheta in radians, could be array
        lorentz site
        """
        x, y = float(self.x), float(self.y)
        self.__hl = x*self.__t_th + y*self.__ic_th


    def _calc_hpveta(self):
        """
        ttheta in radians, could be array
        pseudo-Voight function
        """
        hg = self.__hg
        hl = self.__hl

        hg_2, hl_2 = hg**2, hl**2
        hg_3, hl_3 = hg_2*hg, hl_2*hl
        hg_4, hl_4 = hg_3*hg, hl_3*hl
        hg_5, hl_5 = hg_4*hg, hl_4*hl
        c_2, c_3, c_4, c_5 = 2.69269, 2.42843, 4.47163, 0.07842
        hpv = (hg_5 + c_2*hg_4*hl + c_3*hg_3*hl_2 + 
                       c_4*hg_2*hl_3 + c_5*hg*hl_4 + hl_5)**0.2
        hh = hl*1./hpv
        self.__hpv = hpv 
        self.__eta = 1.36603*hh - 0.47719*hh**2 + 0.11116*hh**3

    def _calc_agbg(self):
        hpv = self.__hpv

        self.__ag = (2./hpv)*(numpy.log(2.)/numpy.pi)**0.5
        self.__bg = 4*numpy.log(2)/(hpv**2)
        
    def _calc_albl(self):
        hpv = self.__hpv
        self.__al = 2./(numpy.pi*hpv )
        self.__bl = 4./(hpv**2)
    
    def calc_resolution(self, tth_hkl, i_g = 0.):
        """
        Calculate parameters for tth
        tth_hkl in degrees
        """
        self._calc_tancos(0.5*tth_hkl*numpy.pi/180.)
        self._calc_hg(i_g = i_g)
        self._calc_hl()
        self._calc_hpveta()
        self._calc_agbg()
        self._calc_albl()

        a_g = self.__ag  
        b_g = self.__bg  
        a_l = self.__al  
        b_l = self.__bl 
        h_g = self.__hg  
        h_l = self.__hl 
        h_pv = self.__hpv 
        eta = self.__eta
        
        return h_pv, eta, h_g, h_l, a_g, b_g, a_l, b_l
    



"""
define class _pd2d_instr_reflex_asymmetry to describe asymmetry of Bragg reflections for 1d powder diffractometer
"""

__author__ = 'ikibalin'
__version__ = "2019_09_10"

import os
import numpy


from pycifstar import Data
from cryspy.f_common.cl_fitable import Fitable



class Pd2dInstrReflexAsymmetry(dict):
    """
    Pd2dInstrReflexAsymmetry describes asymmetry of Bragg reflections for 2d powder diffractometer

    Example:

    _pd2d_instr_reflex_asymmetry_p1 0.0
    _pd2d_instr_reflex_asymmetry_p2 0.0
    _pd2d_instr_reflex_asymmetry_p3 0.0
    _pd2d_instr_reflex_asymmetry_p4 0.0
    """
    def __init__(self, p1 = Fitable(0.), p2 = Fitable(0.), 
                       p3 = Fitable(0.), p4 = Fitable(0.)):
        super(Pd2dInstrReflexAsymmetry, self).__init__()
        self.__pd2d_instr_reflex_asymmetry_p1 = None
        self.__pd2d_instr_reflex_asymmetry_p2 = None
        self.__pd2d_instr_reflex_asymmetry_p3 = None
        self.__pd2d_instr_reflex_asymmetry_p4 = None

        self.p1 = p1
        self.p2 = p2
        self.p3 = p3
        self.p4 = p4

    @property
    def p1(self):
        return self.__pd2d_instr_reflex_asymmetry_p1
    @p1.setter
    def p1(self, x):
        if isinstance(x, Fitable):
            x_in = x
        else:
            x_in = Fitable()
            flag = x_in.take_it(x)
        self.__pd2d_instr_reflex_asymmetry_p1 = x_in

    @property
    def p2(self):
        return self.__pd2d_instr_reflex_asymmetry_p2
    @p2.setter
    def p2(self, x):
        if isinstance(x, Fitable):
            x_in = x
        else:
            x_in = Fitable()
            flag = x_in.take_it(x)
        self.__pd2d_instr_reflex_asymmetry_p2 = x_in

    @property
    def p3(self):
        return self.__pd2d_instr_reflex_asymmetry_p3
    @p3.setter
    def p3(self, x):
        if isinstance(x, Fitable):
            x_in = x
        else:
            x_in = Fitable()
            flag = x_in.take_it(x)
        self.__pd2d_instr_reflex_asymmetry_p3 = x_in

    @property
    def p4(self):
        return self.__pd2d_instr_reflex_asymmetry_p4
    @p4.setter
    def p4(self, x):
        if isinstance(x, Fitable):
            x_in = x
        else:
            x_in = Fitable()
            flag = x_in.take_it(x)
        self.__pd2d_instr_reflex_asymmetry_p4 = x_in


    def __repr__(self):
        ls_out = ["Pd2dInstrResolution:"]
        ls_out.append(" p1: {:}".format(self.p1.print_with_sigma))
        ls_out.append(" p2: {:}".format(self.p2.print_with_sigma))
        ls_out.append(" p3: {:}".format(self.p3.print_with_sigma))
        ls_out.append(" p4: {:}".format(self.p4.print_with_sigma))
        return "\n".join(ls_out)

    @property
    def is_variable(self):
        """
        Output: True if there is any refined parameter
        """
        res = any([self.p1.refinement, 
                   self.p2.refinement,
                   self.p3.refinement,
                   self.p4.refinement])
        return res        
    
    def get_variables(self):
        """
        Output: the list of the refined parameters
        """
        l_variable = []
        if self.p1.refinement:
            l_variable.append(self.p1)
        if self.p2.refinement:
            l_variable.append(self.p2)
        if self.p3.refinement:
            l_variable.append(self.p3)
        if self.p4.refinement:
            l_variable.append(self.p4)
        return l_variable

    def _show_message(self, s_out: str):
        print("***  Error ***")
        print(s_out)
        
    @property
    def to_cif(self):
        ls_out = []
        ls_out.append("_pd2d_instr_reflex_asymmetry_p1 {:}".format(self.p1.print_with_sigma))
        ls_out.append("_pd2d_instr_reflex_asymmetry_p2 {:}".format(self.p2.print_with_sigma))
        ls_out.append("_pd2d_instr_reflex_asymmetry_p3 {:}".format(self.p3.print_with_sigma))
        ls_out.append("_pd2d_instr_reflex_asymmetry_p4 {:}".format(self.p4.print_with_sigma))
        return "\n".join(ls_out)


    def from_cif(self, string: str):
        cif_data = Data()
        flag = cif_data.take_from_string(string)
        if not flag:
            return False
        flag = False
        if cif_data.is_value("_pd2d_instr_reflex_asymmetry_p1"):
            self.p1 = cif_data["_pd2d_instr_reflex_asymmetry_p1"] # CIFvalue
        if cif_data.is_value("_pd2d_instr_reflex_asymmetry_p2"):
            self.p2 = cif_data["_pd2d_instr_reflex_asymmetry_p2"] # CIFvalue
        if cif_data.is_value("_pd2d_instr_reflex_asymmetry_p3"):
            self.p3 = cif_data["_pd2d_instr_reflex_asymmetry_p3"] # CIFvalue
        if cif_data.is_value("_pd2d_instr_reflex_asymmetry_p4"):
            self.p4 = cif_data["_pd2d_instr_reflex_asymmetry_p4"] # CIFvalue
        return True

        
    def _func_fa(self, tth):
        """
        for assymmetry correction
        """ 
        return 2*tth*numpy.exp(-tth**2)
        
    def _func_fb(self, tth):
        """
        for assymmetry correction
        """ 
        return 2.*(2.*tth**2-3.)* self._func_fa(tth)
        
    def calc_asymmetry(self, tth, tth_hkl):
        """
        Calculate asymmetry coefficients for  on the given list ttheta for 
        bragg reflections flaced on the position ttheta_hkl
        tth and tth_hkl in degrees
        
        IMPORTANT: THERE IS MISTAKE (look page 54 in FullProf Manual)
        """
        tth_2d, tth_hkl_2d = numpy.meshgrid(tth, tth_hkl, indexing="ij")
        np_zero = numpy.zeros(tth_2d.shape, dtype = float)
        np_one = numpy.ones(tth_2d.shape, dtype = float)
        val_1, val_2 = np_zero, np_zero
        
        
        p1, p2 = float(self.p1), float(self.p2)
        p3, p4 = float(self.p3), float(self.p4)
        flag_1, flag_2 = False, False
        if ((p1!= 0.)|(p3!= 0.)):
            flag_1 = True
            fa = self._func_fa(tth)
        if ((p2!= 0.)|(p4!= 0.)):
            flag_2 = True
            fb = self._func_fb(tth)
            
        flag_3, flag_4 = False, False
        if ((p1!= 0.)|(p2!= 0.)):
            if flag_1:
                val_1 += p1*fa
                flag_3 = True
            if flag_2:
                val_1 += p2*fb
                flag_3 = True
            if flag_3:
                c1 = 1./numpy.tanh(0.5*tth_hkl)
                c1_2d = numpy.meshgrid(tth, c1, indexing="ij")[1]
                val_1 *= c1_2d

        if ((p3!= 0.)|(p4!= 0.)):
            if flag_1:
                val_2 += p3*fa
                flag_4 = True
            if flag_2:
                val_2 += p4*fb
                flag_4 = True
            if flag_4:
                c2 = 1./numpy.tanh(tth_hkl)
                c2_2d = numpy.meshgrid(tth, c2, indexing="ij")[1]
                val_2 *= c2_2d

        asymmetry_2d = np_one+val_1+val_2
        return asymmetry_2d
    
